import typing as t

import numpy as np

"""
This is a specific list consisting of exactly three float values in the range [0, 1]. The entire list thus
represents an RGB color value, where the corresponding elements denote the intensity of the colors
red green and blue respectively.
"""
RgbList = t.Union[t.List[float], t.Tuple[float, float, float]]

"""
This is an alias which specifies the main graph representation which is used throughout this application.
A graph is generally represented as a dict structure with very key value pairs.

The following keys are always expected to be present:

- ``node_indices``: A (N, ) array of integers representing the unique node indices
- ``node_attributes``: A (N, K) array of float node features
- ``node_adjacency``: A (N, N) array which is the adjacency matrix for the graph. (Binary entries)
- ``edge_indices``: A (M, 2) array, which is essentially a list of all edges (specified as two int node
  indices) without any particular order
- ``edge_attributes``: A (M, ) array of edge weights

The following keys are *optionally* present / expected for certain tasks:

- ``node_positions``: A (N, 2) array of float positions for the nodes on a 2D plane. These are used to place
  the nodes into a matplotlib coordinate system.
- ``node_coordinates``: A (N, 2) array of float positions for the nodes within a corresponding *image*. These
  values are the pixel coordinates in the final image!
- ``node_importances``: A (N, ) array of float values determining node importance values (=explanations)
- ``edge_importances``: A (M, ) array of float values determining edge importance values (=explanations). The
  values are assigned to the edges as they are ordered in edge_indices!
"""
GraphDict = t.Dict[str, np.ndarray]

"""
A dictionary which contains the metadata corresponding to an element of the dataset. This metadata dict 
contains all the relevant information about the representation of a graph as well as additional information 
about the prediction target for example.

There are some keys which all the metadata dicts should have in common, but theoretically any additional 
fields can be added for special purpose annotations of different datasets.
"""
MetadataDict = t.Dict[str, t.Union[GraphDict, str, float, t.Any]]

"""
This is a special dictionary structure which represents a loaded "eye tracking dataset". An eye tracking 
dataset is a special format for a graph dataset, where the whole dataset is represented by a folder in the 
file system. Within this folder, every element is represented by two files: A JSON file which contains the 
graph representation and the metadata and a PNG file which visualizes the graph in a domain specific way.

The keys of this dict are the *names* of the elements. Those are also used as the file names for both the 
JSON and the PNG files in the folder structure. The corresponding values are again dicts which contain the 
following fields:

- ``index``: An integer for the order in which the elements were loaded from the filesystem. NOTE that this 
  index may differ between operating systems and dev environments as it depends on how the files are 
  sorted in the folder.
- ``name``: A string for the name of the element, which is determined by the file name
- ``image_path``: The absolute string path to the image file that contains the visualization of the 
  corresponding graph element
- ``metadata``: A dictionary which contains the metadata about the element as well as the graph 
  representation of the element itself. Metadata for example contains information about the target value 
  of the element and a possible dataset split as well as usually a canonical index within the dataset
"""
EtDatasetDict = t.Dict[str, t.Union[str, dict]]



import os
import pathlib
import random
import typing as t
from collections import Counter

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import roc_auc_score
from pycomex.util import Skippable
from pycomex.experiment import SubExperiment
from visual_graph_datasets.data import load_visual_graph_dataset
from visual_graph_datasets.visualization.importances import create_importances_pdf
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.layers.modules import DropoutEmbedding, DenseEmbedding, LazyConcatenate
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.pooling import PoolingGlobalEdges

import graph_attention_student.typing as tc
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.models import gnnx_importances
from graph_attention_student.training import mse, NoLoss
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.util import array_normalize, binary_threshold

# == DATASET PARAMETERS ==
# These parameters define the dataset that is to be used for the experiment as well as properties of
# that dataset such as the train test split for example.

# The number of importance channels reflected in the ground truth importances loaded from the dataset
IMPORTANCE_CHANNELS = 1
# The string key of the graph dict representations to use to retrieve the ground truth node importances
# ! Note: A visual graph dataset may have multiple different ground truth explanation annotations with
# different numbers of channels!
NODE_IMPORTANCES_KEY: t.Optional[str] = 'node_importances_1'
# The string key of the graph dict representations to use to retrieve the ground truth edge importances
EDGE_IMPORTANCES_KEY: t.Optional[str] = 'edge_importances_1'

# If this flag is true, the "node_coordinates" will be added to the other node attributes
USE_NODE_COORDINATES: bool = False

# If this flag is true, the "edge_lengths" will be added to the other edge attributes
USE_EDGE_LENGTHS: bool = False

# In this default implementation we use a GCN network, which cannot handle edge attributes!
USE_EDGE_ATTRIBUTES: bool = False

# == MODEL PARAMETERS ==

MODEL_NAME = 'GNNX+GCN'

class GcnModel(ks.models.Model):

    def __init__(self,
                 units: t.List[int],
                 final_units: t.List[int],
                 pooling_method: str = 'sum',
                 activation: str = 'kgcnn>leaky_relu',
                 final_pooling: str = 'sum',
                 final_activation: str = 'linear',
                 dropout_rate: float = 0.0):
        super(GcnModel, self).__init__()

        self.lay_dropout = DropoutEmbedding(rate=dropout_rate)
        self.conv_layers = []
        for k in units:
            lay = GCN(units=k, pooling_method=pooling_method, activation=activation)
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingGlobalEdges(pooling_method=final_pooling)
        self.lay_concat = LazyConcatenate(axis=-1)

        self.final_layers = []
        self.final_activations = ['relu' for _ in final_units]
        self.final_activations[-1] = final_activation
        for k, act in zip(final_units, self.final_activations):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call(self, inputs, training=False):
        node_input, edge_input, edge_index_input = inputs

        x = node_input
        for lay in self.conv_layers:
            x = lay([x, edge_input, edge_index_input])
            if training:
                x = self.lay_dropout(x)

        final = self.lay_pooling(x)
        for lay in self.final_layers:
            final = lay(final)

        return final


UNITS = [32, 32, 32]
FINAL_UNITS = [32, 16, 1]
DROPOUT_RATE = 0.0

# == TRAINING PARAMETERS ==
# These parameters control the training process of the neural network.

# The number of independent training process repetitions to get a statistical measure of the performance
REPETITIONS: int = 1
# This optimizer will be used during training
OPTIMIZER: ks.optimizers.Optimizer = ks.optimizers.Nadam(learning_rate=0.01)
BATCH_SIZE: int = 256
EPOCHS: int = 250
DEVICE: str = 'gpu:0'

# == GNNX PARAMETERS ==
# Parameters for the GnnExplainer training to create the explanations

GNNX_EPOCHS = 500
GNNX_LEARNING_RATE = 0.02
GNNX_NODE_SPARSITY = 0.3
GNNX_EDGE_SPARSITY = 0.3

# == EVALUATION PARAMETERS ==
# These parameters control the evaluation process, which includes the drawing of visualizations and plots

# After how many elements a log step is printed to the console
LOG_STEP_EVAL: int = 100
# This is the batch size that is used during the evaluation of the test set.
BATCH_SIZE_EVAL: int = 256

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_single.py')
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_gnnx'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    @se.hook('create_model')
    def create_model(e):
        model = GcnModel(
            units=e.parameters['UNITS'],
            final_units=e.parameters['FINAL_UNITS'],
            dropout_rate=e.parameters['DROPOUT_RATE']
        )
        model.compile(
            loss=mse,
            metrics=mse,
            optimizer=e.parameters['OPTIMIZER_CB'](),
            run_eagerly=False
        )
        return model

    @se.hook('fit_model')
    def fit_model(e, model, x_train, y_train, x_test, y_test):
        history = model.fit(
            x_train,
            y_train,
            epochs=e.parameters['EPOCHS'],
            batch_size=e.parameters['BATCH_SIZE'],
            validation_data=(x_test, y_test),
            validation_freq=1,
            callbacks=[
                LogProgressCallback(
                    e.logger,
                    epoch_step=5,
                    identifier='val_mean_squared_error'
                )
            ],
            verbose=0
        )
        return history.history

    @se.hook('query_model')
    def query_model(e, model, x, y, include_importances: bool = True):
        e.info('querying the model...')
        out_pred = model(x)

        if include_importances:
            e.info('creating GNNX explanations...')
            ni_gnnx, ei_gnnx = gnnx_importances(
                model,
                x=x,
                y=out_pred,
                node_sparsity_factor=e.parameters['GNNX_NODE_SPARSITY'],
                edge_sparsity_factor=e.parameters['GNNX_EDGE_SPARSITY'],
                epochs=e.parameters['GNNX_EPOCHS'],
                learning_rate=e.parameters['GNNX_LEARNING_RATE'],
                logger=e.logger
            )
            return out_pred, ni_gnnx, ei_gnnx

        else:
            return out_pred

    @se.hook('calculate_fidelity')
    def calculate_fidelity(e, model, indices_true, x_true, y_true, out_pred, ni_pred, ei_pred):
        node_masks = []
        edge_masks = []
        for c, (ni, ei) in enumerate(zip(ni_pred, ei_pred)):
            node_mask = np.ones_like(ni)
            node_mask[array_normalize(ni) > 0.5] = 0
            node_masks.append(node_mask)

            edge_mask = np.ones_like(ei)
            edge_mask[array_normalize(ei) > 0.5] = 0
            edge_masks.append(edge_mask)

        node_masks_tensor = ragged_tensor_from_nested_numpy(node_masks)
        edge_masks_tensor = ragged_tensor_from_nested_numpy(edge_masks)

        x_masked = (
            x_true[0] * node_masks_tensor,
            x_true[1] * edge_masks_tensor,
            *x_true[2:]
        )
        out_masked = model(x_masked)
        out_masked = out_masked.numpy()

        fidelities = []
        for c, (value_pred, value_masked) in enumerate(zip(out_pred, out_masked)):
            fidelity = np.abs(value_pred - value_masked)
            fidelities.append(fidelity)

        return fidelities




import typing as t

import tensorflow as tf
import tensorflow.keras as ks
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.modules import DenseEmbedding
from kgcnn.layers.pooling import PoolingNodes

from graph_attention_student.layers import ExplanationSparsityRegularization
from graph_attention_student.models.gradient import AbstractGradientModel


class GnesGradientModel(AbstractGradientModel):

    def __init__(self,
                 units: t.List[int],
                 batch_size: int,
                 importance_func: t.Callable,
                 layer_cb: t.Callable = lambda units: GCN(units=units, activation='kgcnn>leaky_relu'),
                 sparsity_factor: float = 0.1,
                 final_units: t.List[int] = [1],
                 final_activation: str = 'linear',
                 pooling_method: str = 'sum'):
        super(GnesGradientModel, self).__init__()
        self.importance_func = importance_func
        self.batch_size = batch_size
        self.num_outputs = final_units[-1]

        self.conv_layers = []
        for k in units:
            lay = layer_cb(k)
            self.conv_layers.append(lay)

        # ~ global pooling
        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)

        # ~ mlp tail-end
        self.final_units = final_units
        self.final_acts = ['kgcnn>leaky_relu' for _ in self.final_units]
        self.final_acts[-1] = final_activation
        self.final_layers = []
        for k, act in zip(self.final_units, self.final_acts):
            lay = DenseEmbedding(
                units=k,
                activation=act
            )
            self.final_layers.append(lay)

        # ~ regularization
        self.lay_sparsity = ExplanationSparsityRegularization(factor=sparsity_factor)

    def call_with_gradients(self, inputs, training=True, batch_size=None, create_gradients=True):
        if batch_size is None:
            batch_size = self.batch_size

        node_input, edge_input, edge_index_input = inputs
        x = node_input

        with tf.GradientTape(persistent=True) as tape:
            tape.watch(node_input)
            tape.watch(edge_input)

            # ~ convolutional layers
            edge_activations = [edge_input]
            node_activations = [node_input]
            for lay in self.conv_layers:
                x = lay([x, edge_input, edge_index_input])
                node_activations.append(x)

            # ~ global pooling
            out = self.lay_pooling(x)

            # ~ mlp tail end
            for lay in self.final_layers:
                out = lay(out)

            outs = [[out[b, o] for o in range(self.num_outputs)] for b in range(batch_size)]

        if create_gradients:
            node_gradient_info = self.calculate_gradient_info(outs, node_activations, tape, batch_size)
            edge_gradient_info = self.calculate_gradient_info(outs, edge_activations, tape, batch_size)

            node_importances = self.importance_func(node_gradient_info)
            edge_importances = self.importance_func(edge_gradient_info)

            self.lay_sparsity(node_importances)
            self.lay_sparsity(edge_importances)

            return out, node_importances, edge_importances

        else:
            return out, None, None

    def call(self, inputs, *args, **kwargs):
        return self.call_with_gradients(inputs, *args, **kwargs)


"""

"""
import os
import pathlib
import random
import warnings
import typing as t

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from pycomex.util import Skippable
from pycomex.experiment import Experiment
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import r2_score
from sklearn.metrics import roc_auc_score
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from visual_graph_datasets.data import load_visual_graph_dataset
from visual_graph_datasets.visualization.importances import create_importances_pdf
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.models import GcnGradientModel
from graph_attention_student.models import grad_importances
from graph_attention_student.models import gnnx_importances
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.training import mae
from graph_attention_student.util import array_normalize
from graph_attention_student.util import binary_threshold
from graph_attention_student.util import render_latex
from graph_attention_student.util import latex_table
from graph_attention_student.util import latex_table_element_mean

warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
PATH = pathlib.Path(__file__).parent.absolute()

# == DATASET PARAMETERS ==
VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/rb_dual_motifs')
RATIO_DATASET: t.Optional[float] = None
RATIO_TRAIN: float = 0.8
NODE_IMPORTANCES_KEY = 'node_importances_1'
EDGE_IMPORTANCES_KEY = 'edge_importances_1'
NUM_EXAMPLE: int = 100

# == TRAINING PARAMETERS ==
REPETITIONS = 25
DEVICE = 'cpu:0'
BATCH_SIZE = 32
OPTIMIZER_CB = lambda: ks.optimizers.Adam(learning_rate=0.01)
EPOCHS = 25

# == MODEL PARAMETERS ==
UNITS: t.List[int] = [12, 12, 12]
FINAL_UNITS: t.List[int] = [6, 1]
FINAL_ACTIVATION: str = 'linear'

# == EVALUATION PARAMETERS ==
# :param NUM_RANDOM_FIDELITY: This integer number determines the number of times which the "random fidelity"
#       is going to be sampled for each element of the dataset.
NUM_RANDOM_FIDELITY: int = 10
GNNX_EPOCHS: int = 200
GNNX_SPARSITY_FACTOR: float = 1.0

# == EXPERIMENT PARAMETERS ==
BASE_PATH = os.path.join(PATH, 'results')
NAMESPACE = os.path.basename(__file__).strip('.py')
DEBUG = True
TESTING = False

# If the TESTING flag is set that means that an execution of the experiment is supposed to be only a
# test of functionality. In that case we want to modify several parameters such that the length of the
# experiment is drastically reduced.
if TESTING:
    RATIO_DATASET = 0.2
    REPETITIONS = 1
    EPOCHS = 10
    NUM_RANDOM_MASKS = 1
    GNNX_EPOCHS = 10

with Skippable(), (e := Experiment(BASE_PATH, NAMESPACE, globals())):
    vgd_name = os.path.basename(VISUAL_GRAPH_DATASET_PATH)
    e.info(f'using visual graph dataset "{vgd_name}"')

    # ~ loading the dataset
    e.info('loading visual graph dataset from disk')
    metadata_map, index_data_map = load_visual_graph_dataset(
        VISUAL_GRAPH_DATASET_PATH,
        logger=e.logger,
        log_step=1000,
        metadata_contains_index=True,
    )
    dataset_indices_all = list(sorted(index_data_map.keys()))
    dataset_length = len(dataset_indices_all)

    dataset: t.List[dict] = [None for _ in dataset_indices_all]
    for index in dataset_indices_all:
        g = index_data_map[index]['metadata']['graph']

        g['node_importances'] = g[NODE_IMPORTANCES_KEY]
        g['edge_importances'] = g[EDGE_IMPORTANCES_KEY]

        dataset[index] = g

    e.info(f'loaded dataset with {len(dataset)} elements')


    @e.hook('explanation_methods', default=True)
    def explanation_methods(_e,
                            model: ks.models.Model,
                            test_indices: t.List[int],
                            x_test: tuple,
                            y_test: tuple):

        # ~ Simple Gradient Explanations
        out_pred, node_gradient_info, edge_gradient_info = model(
            x_test,
            return_gradients=True,
            batch_size=len(test_indices),
        )
        ni_grad = grad_importances(
            gradient_info=node_gradient_info,
            use_absolute=True,
            keepdims=False,
        )
        ei_grad = grad_importances(
            gradient_info=edge_gradient_info,
            use_absolute=True,
            keepdims=False,
        )
        ni_grad = ni_grad.numpy()
        ei_grad = ei_grad.numpy()

        yield 'grad', ni_grad, ei_grad

        # ~ GNNExplainer
        ni_gnnx, ei_gnnx = gnnx_importances(
            model=lambda x: model(x, create_gradients=False, return_gradients=False),
            x=x_test,
            y=out_pred,
            epochs=_e.parameters['GNNX_EPOCHS'],
            node_sparsity_factor=_e.parameters['GNNX_SPARSITY_FACTOR'],
            edge_sparsity_factor=_e.parameters['GNNX_SPARSITY_FACTOR'],
            logger=_e.logger,
            log_step=10
        )
        ni_gnnx = ni_gnnx.numpy()
        ei_gnnx = ei_gnnx.numpy()

        yield 'gnnx', ni_gnnx, ei_gnnx

    e.info(f'starting explanations')
    for rep in range(REPETITIONS):
        e.info(f'REPETITION ({rep+1}/{REPETITIONS})')

        with tf.device(DEVICE):
            e.info('creating repetition archive folder...')
            rep_path = os.path.join(e.path, f'{rep:02d}')
            os.mkdir(rep_path)

            e.info('creating random dataset split...')
            dataset_indices = dataset_indices_all.copy()
            if RATIO_DATASET:
                num_elements = int(RATIO_DATASET * dataset_length)
                dataset_indices = dataset_indices[:num_elements]

            dataset_indices_set = set(dataset_indices)
            num_train = int(len(dataset_indices) * RATIO_TRAIN)
            train_indices = random.sample(dataset_indices, k=num_train)
            train_indices_set = set(train_indices)
            test_indices_set = dataset_indices_set.difference(train_indices_set)
            test_indices = list(test_indices_set)

            num_example = min(len(test_indices), NUM_EXAMPLE)
            example_indices = random.sample(test_indices, k=num_example)

            e[f'train_indices/{rep}'] = train_indices
            e[f'test_indices/{rep}'] = test_indices
            e[f'example_indices/{rep}'] = example_indices
            e.info(f'created split with '
                   f'{len(train_indices)} train elements, '
                   f'{len(test_indices)} test elements, '
                   f'{len(example_indices)} example elements for visualization')

            e.info('processing the dataset into tensors...')
            x_train, y_train, x_test, y_test = process_graph_dataset(
                dataset,
                test_indices=test_indices,
                train_indices=train_indices,
                use_importances=True,
                use_graph_attributes=False,
            )

            e.info('creating the model...')
            model: ks.Model = GcnGradientModel(
                batch_size=BATCH_SIZE,
                units=UNITS,
                final_units=FINAL_UNITS,
                final_activation=FINAL_ACTIVATION,
                layer_cb=lambda units: AttentionHeadGATV2(
                    units=units,
                    use_edge_features=True,
                ),
                pooling_method='sum',
            )
            model.compile(
                loss=[ks.losses.MeanSquaredError()],
                loss_weights=[1],
                metrics=[ks.metrics.MeanSquaredError()],
                optimizer=OPTIMIZER_CB(),
                run_eagerly=False,
            )

            e.info('starting model training...')
            hist = model.fit(
                x_train,
                y_train,
                batch_size=BATCH_SIZE,
                epochs=EPOCHS,
                callbacks=[LogProgressCallback(
                    logger=e.logger,
                    epoch_step=1,
                    identifier=f'loss'
                )],
                verbose=0,
            )
            history = hist.history
            e[f'history/{rep}'] = history
            e.info(f'finished training for model with {model.count_params()} parameters')

            e.info('evaluating on test set...')

            # ~ The overall prediction metrics
            out_pred, node_gradient_info, edge_gradient_info = model(x_test, return_gradients=True)
            out_pred = out_pred.numpy()
            out_true, ni_true, ei_true = [v.numpy() if not isinstance(v, np.ndarray) else v
                                          for v in y_test]

            mse = mean_squared_error(out_true, out_pred)
            mae = mean_absolute_error(out_true, out_pred)
            r2 = r2_score(out_true, out_pred)
            e.info(f'prediction metrics:\n'
                   f' * mse: {mse:.3f}\n'
                   f' * mae: {mae:.3f}\n'
                   f' * r2: {r2:.3f}\n')

            e[f'mse/{rep}'] = mse
            e[f'mae/{rep}'] = mae
            e[f'r2/{rep}'] = r2

            for c, index in enumerate(test_indices):
                e[f'out/true/{rep}/{index}'] = out_true[c]
                e[f'out/pred/{rep}/{index}'] = out_pred[c]

            e['keys'] = []
            for key, ni_pred, ei_pred in e.apply_hook('explanation_methods',
                                                      model=model,
                                                      test_indices=test_indices,
                                                      x_test=x_test,
                                                      y_test=y_test):
                e.info(f'explanation method "{key}"...')
                e['keys'].append(key)

                node_masks = []
                edge_masks = []
                for c, index in enumerate(test_indices):
                    ni = array_normalize(ni_pred[c])
                    ei = array_normalize(ei_pred[c])
                    e[f'ni/{key}/{rep}/{index}'] = ni
                    e[f'ei/{key}/{rep}/{index}'] = ei

                    # Up until this point, every importance value is still a float, but we need to the binary
                    # explanations for two reasons: To construct the masks for the fidelity computation and
                    # for the computation of the sparsity
                    ni_binary = binary_threshold(ni, threshold=0.5)
                    ei_binary = binary_threshold(ei, threshold=0.5)
                    node_masks.append(1.0 - ni_binary.astype(float))
                    edge_masks.append(1.0 - ei_binary.astype(float))

                    node_sparsity = np.mean(ni_binary)
                    edge_sparsity = np.mean(ei_binary)
                    e[f'node_sparsity/{key}/{rep}/{index}'] = node_sparsity
                    e[f'edge_sparsity/{key}/{rep}/{index}'] = edge_sparsity

                e.info('calculating fidelity...')
                node_mask_tensor = ragged_tensor_from_nested_numpy(node_masks)
                edge_mask_tensor = ragged_tensor_from_nested_numpy(edge_masks)
                out_masked = model((
                    x_test[0] * node_mask_tensor,
                    x_test[1] * edge_mask_tensor,
                    x_test[2]
                ), create_gradients=False, return_gradients=False)
                for c, index in enumerate(test_indices):
                    # Here we calculate the fidelity as the absolute difference in the predicted value.
                    # We multiply this with the fidelity of that element to make it a bit more comparable for
                    # the following reason: An explanation which consists of more explanations (and thus a
                    # bigger mask for the computation of the sparsity) will likely result in a bigger
                    # deviation naturally because more input elements are being perturbed.
                    fidelity = abs(out_masked[c] - out_pred[c])  # * e[f'node_sparsity/{key}/{rep}/{index}']
                    e[f'fidelity/{key}/{rep}/{index}'] = fidelity

                e.info('calculating random fidelity...')
                for j in range(NUM_RANDOM_FIDELITY):
                    # first of all we need to create the random masks, which we can very easily do by simply
                    # using the "correct" explanation masks we already have and simply randomly permutating
                    # them, because one important property for the random masks is that they have to have
                    # the same sparsity!
                    for node_mask, edge_mask in zip(node_masks, edge_masks):
                        np.random.shuffle(node_mask)
                        np.random.shuffle(edge_mask)

                    node_mask_tensor = ragged_tensor_from_nested_numpy(node_masks)
                    edge_mask_tensor = ragged_tensor_from_nested_numpy(edge_masks)

                    out_masked = model((
                        x_test[0] * node_mask_tensor,
                        x_test[1] * edge_mask_tensor,
                        x_test[2]
                    ), create_gradients=False, return_gradients=False)
                    for c, index in enumerate(test_indices):
                        fidelity = abs(out_masked[c] - out_pred[c])  # * e[f'node_sparsity/{key}/{rep}/{index}']
                        e[f'random_fidelity/{key}/{rep}/{index}/{j}'] = fidelity

                # ~ calculating the explanation specific metrics
                node_auc = roc_auc_score(
                    np.concatenate([v.flatten() for v in ni_true]),
                    np.concatenate([v.flatten() for v in ni_pred]),
                )
                edge_auc = roc_auc_score(
                    np.concatenate([v.flatten() for v in ei_true]),
                    np.concatenate([v.flatten() for v in ei_pred]),
                )
                e[f'node_auc/{key}/{rep}'] = node_auc
                e[f'edge_auc/{key}/{rep}'] = edge_auc

                node_sparsities = [e[f'node_sparsity/{key}/{rep}/{index}'] for index in test_indices]
                edge_sparsities = [e[f'edge_sparsity/{key}/{rep}/{index}'] for index in test_indices]

                fidelities = np.array([e[f'fidelity/{key}/{rep}/{index}'] for index in test_indices])
                fidelity_mean = np.mean(fidelities)
                fidelity_std = np.std(fidelities)
                random_fidelities = np.array([np.mean([e[f'random_fidelity/{key}/{rep}/{index}/{j}']
                                                       for j in range(NUM_RANDOM_FIDELITY)])
                                              for index in test_indices])
                diff_fidelities = fidelities - random_fidelities
                for c, index in enumerate(test_indices):
                    e[f'rand_fidelity/{key}/{rep}/{index}'] = random_fidelities[c]
                    e[f'diff_fidelity/{key}/{rep}/{index}'] = diff_fidelities[c]

                e.info(f'explanation metrics:\n'
                       f' * node auc: {node_auc:.3f}\n'
                       f' * edge auc: {edge_auc:.3f}\n'
                       f' * mean node sparsity: {np.mean(node_sparsities):.3f}\n'
                       f' * mean edge sparsity: {np.mean(edge_sparsities):.3f}\n'
                       f' * mean fidelity: {fidelity_mean:.3f}\n'
                       f' * std fidelity: {fidelity_std:.3f}\n'
                       f' * mean random fidelity: {np.mean(random_fidelities):.3f}\n'
                       f' * mean diff fidelity: {np.mean(diff_fidelities):.3f}\n'
                       )

            e.info('visualizing examples...')
            graph_list = [index_data_map[i]['metadata']['graph'] for i in example_indices]
            image_path_list = [index_data_map[i]['image_path'] for i in example_indices]
            node_positions_list = [g['node_positions'] for g in graph_list]
            output_path = os.path.join(rep_path, 'examples.pdf')
            create_importances_pdf(
                graph_list=graph_list,
                image_path_list=image_path_list,
                node_positions_list=node_positions_list,
                importances_map={
                    'gt': (
                        [g['node_importances'] for g in graph_list],
                        [g['edge_importances'] for g in graph_list],
                    ),
                    **{
                        key: (
                            [e[f'ni/{key}/{rep}/{index}'] for index in example_indices],
                            [e[f'ei/{key}/{rep}/{index}'] for index in example_indices]
                        )
                        for key in e['keys']
                    }
                },
                output_path=output_path,
                logger=e.logger,
                log_step=100,
            )


with Skippable(), e.analysis:

    e.info('starting analysis of experiment results...')

    # ~ Latex table with the results over all repetitions
    e.info('rendering latex table with the results...')
    column_names = [
        r'Model Key',
        r'$\text{MSE} \downarrow$',
        r'$r^{2} \uparrow$',
        r'$\text{Node AUC} \uparrow$',
        r'$\text{Edge AUC} \uparrow$',
        r'$\text{Node Sparsity} \downarrow$',
        r'$\text{Edge Sparsity} \downarrow$',
        r'$\text{Fidelity} \uparrow$',
    ]
    rows = []
    for key in e['keys']:
        row = []
        row.append(key)
        row.append([e[f'mse/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([e[f'r2/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([e[f'node_auc/{key}/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([e[f'edge_auc/{key}/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([np.mean(list(e[f'node_sparsity/{key}/{rep}'].values()))
                    for rep in range(REPETITIONS)])
        row.append([np.mean(list(e[f'edge_sparsity/{key}/{rep}'].values()))
                    for rep in range(REPETITIONS)])
        row.append([np.mean(list(e[f'diff_fidelity/{key}/{rep}'].values()))
                    for rep in range(REPETITIONS)])

        rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_mean,
    )
    e.commit_raw('table.tex', table)
    output_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=output_path)
    e.info('rendered latex table with the results')



import os
import sys
import pathlib
import random
import typing as t
from collections import Counter

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import roc_auc_score
from pycomex.util import Skippable
from pycomex.experiment import Experiment
from visual_graph_datasets.data import load_visual_graph_dataset
from visual_graph_datasets.visualization.importances import create_importances_pdf
from kgcnn.data.utils import ragged_tensor_from_nested_numpy

import graph_attention_student.typing as tc
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.util import array_normalize, binary_threshold
from graph_attention_student.util import latex_table, latex_table_element_mean
from graph_attention_student.util import render_latex

# == DATASET PARAMETERS ==
# These parameters define the dataset that is to be used for the experiment as well as properties of
# that dataset such as the train test split for example.

# The name of the visual graph dataset to use for this experiment.
VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/rb_dual_motifs')

# If this is None, then a new random split will be created for every repetition of the experiment.
# Otherwise this may be the int index of an existing split that is associated with the dataset and then
# that split will be used instead all the time.
USE_DATASET_SPLIT: t.Optional[int] = None

# The percentage of the dataset to use for random train set
TRAIN_RATIO: float = 0.8

# THis is the number of elements from the test set which is going to be used as examples for the
# visualization of the explanations generated by the model
NUM_EXAMPLES: int = 100
# This may be a list of dataset indices. The corresponding elements will always be part of the example
# set and be visualized
EXAMPLE_INDICES: t.List[int] = []
# The number of importance channels reflected in the ground truth importances loaded from the dataset
IMPORTANCE_CHANNELS = 1
# The string key of the graph dict representations to use to retrieve the ground truth node importances
# ! Note: A visual graph dataset may have multiple different ground truth explanation annotations with
# different numbers of channels!
NODE_IMPORTANCES_KEY: t.Optional[str] = 'node_importances_1'
# The string key of the graph dict representations to use to retrieve the ground truth edge importances
EDGE_IMPORTANCES_KEY: t.Optional[str] = 'edge_importances_1'

# If this flag is true, the "node_coordinates" will be added to the other node attributes
USE_NODE_COORDINATES: bool = False

# If this flag is False, instead of using the "edge_attributes" of the dataset, constant edge weights of
# 1 are going to be used as a substitute (some GNNs cannot handle edge attributes!)
USE_EDGE_ATTRIBUTES: bool = True

# If this flag is true, the "edge_lengths" will be added to the other edge attributes
USE_EDGE_LENGTHS: bool = False


# == MODEL PARAMETERS ==
# These parameters can be used to configure the model
MODEL_NAME: str = 'MOCK'


# == TRAINING PARAMETERS ==
# These parameters control the training process of the neural network.

# The number of independent training process repetitions to get a statistical measure of the performance
REPETITIONS: int = 1
# This optimizer will be used during training
OPTIMIZER_CB = lambda: ks.optimizers.Nadam(learning_rate=0.01)
BATCH_SIZE: int = 256
EPOCHS: int = 200
DEVICE: str = 'gpu:0'

# == EVALUATION PARAMETERS ==
# These parameters control the evaluation process, which includes the drawing of visualizations and plots

# After how many elements a log step is printed to the console
LOG_STEP_EVAL: int = 100
# This is the batch size that is used during the evaluation of the test set.
BATCH_SIZE_EVAL: int = 256
# A list which has to have the same number of entries as there are importance channels in this experiment.
# Each string element defines the label that is to be given to that channel in the visualization of
# the example explanations generated by the model.
IMPORTANCE_CHANNEL_LABELS = ['predicted']

# == EXPERIMENT PARAMETERS ==
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_single'
DEBUG = True
with Skippable(), (e := Experiment(BASE_PATH, NAMESPACE, globals())):

    vgd_name = os.path.basename(VISUAL_GRAPH_DATASET_PATH)
    e.info(f'Starting to train GNNX model on vgd "{vgd_name}"')

    # -- DATASET LOADING --

    metadata_map, index_data_map = load_visual_graph_dataset(
        VISUAL_GRAPH_DATASET_PATH,
        logger=e.logger,
        log_step=LOG_STEP_EVAL,
        metadata_contains_index=True
    )

    # index_data_map = {int(index): data for index, data in index_data_map.items()}
    dataset_indices = list(sorted(index_data_map.keys()))
    dataset_indices_set = set(dataset_indices)
    dataset_length = len(index_data_map)

    # As the first step in the processing pipeline we need to get a list of graph dicts, which can then
    # later be turned into the tensors needed for the model.
    dataset: t.List[tc.GraphDict] = []
    for index in dataset_indices:
        data = index_data_map[index]
        g = data['metadata']['graph']

        if NODE_IMPORTANCES_KEY is not None:
            g['node_importances'] = g[NODE_IMPORTANCES_KEY]
        else:
            g['node_importances'] = np.zeros(shape=(len(g['node_indices']), IMPORTANCE_CHANNELS))

        if EDGE_IMPORTANCES_KEY is not None:
            g['edge_importances'] = g[EDGE_IMPORTANCES_KEY]
        else:
            g['edge_importances'] = np.zeros(shape=(len(g['edge_indices']), IMPORTANCE_CHANNELS))

        if USE_NODE_COORDINATES:
            g['node_attributes'] = np.array([g['node_attributes'][i].tolist() + g['node_coordinates'][i].tolist()
                                             for i in g['node_indices']])

        if USE_EDGE_LENGTHS:
            g['edge_attributes'] = np.array([g['edge_attributes'][i].tolist() + [g['edge_lengths'][i]]
                                             for i, _ in enumerate(g['edge_indices'])])

        if not USE_EDGE_ATTRIBUTES:
            g['edge_attributes'] = np.ones(shape=(len(g['edge_indices']), 1))

        dataset.append(g)

    e.info(f'loaded dataset with {len(dataset)} elements')

    # -- MODEL TRAINING --

    @e.hook('process_dataset', default=True)
    def process_dataset(_e, dataset, train_indices, test_indices):
        x_train, y_train, x_test, y_test = process_graph_dataset(
            dataset,
            train_indices=train_indices,
            test_indices=test_indices,
        )
        return x_train, y_train, x_test, y_test


    for rep in range(REPETITIONS):
        e.info(f'REPETITION ({rep+1}/{REPETITIONS})')
        e['rep'] = rep

        with tf.device(DEVICE):

            if USE_DATASET_SPLIT is not None:
                e.info(f'creating train test split from dataset...')
                train_indices = [index for index, data in index_data_map.items()
                                 if USE_DATASET_SPLIT in data['metadata']['train_indices']]
                test_indices = [index for index, data in index_data_map.items()
                                if USE_DATASET_SPLIT in data['metadata']['test_indices']]

            else:
                e.info(f'creating random train test split...')
                train_indices = random.sample(dataset_indices, k=int(TRAIN_RATIO * dataset_length))
                train_indices_set = set(train_indices)
                test_indices_set =dataset_indices_set.difference(train_indices_set)
                test_indices = list(test_indices_set)

            test_indices_set = set(test_indices) | set(EXAMPLE_INDICES)
            test_indices = list(test_indices_set)
            e.info(f'using {len(train_indices)} training elements and {len(test_indices)} test elements')
            e[f'train_indices/{rep}'] = train_indices
            e[f'test_indices/{rep}'] = test_indices

            num_examples = min(NUM_EXAMPLES, len(test_indices)) - len(EXAMPLE_INDICES)
            example_indices = EXAMPLE_INDICES + list(random.sample(test_indices, k=num_examples))
            e.info(f'using {len(example_indices)} as examples')
            e[f'example_indices/{rep}'] = example_indices

            e.info('converting dataset into ragged tensors for training...')
            x_train, y_train, x_test, y_test = e.apply_hook(
                'process_dataset',
                dataset=dataset,
                train_indices=train_indices,
                test_indices=test_indices
            )

            e.info('creating the model...')
            model: ks.models.Model = e.apply_hook('create_model')

            e.info('fitting the model...')
            history: dict = e.apply_hook(
                'fit_model',
                model=model,
                x_train=x_train,
                y_train=y_train,
                x_test=x_test,
                y_test=y_test,
            )
            e[f'history/{rep}'] = history
            e[f'epochs/{rep}'] = list(range(EPOCHS))

            # -- EVALUATING THE MODEL --
            e.info('evaluating test set...')
            out_pred, ni_pred, ei_pred = e.apply_hook(
                'query_model',
                model=model,
                x=x_test,
                y=y_test,
                include_importances=True,
            )
            out_pred = np.array(out_pred.numpy())
            ni_pred = np.array(ni_pred.numpy())
            ei_pred = np.array(ei_pred.numpy())

            out_true = np.array(y_test[0])
            ni_true = np.array(y_test[1].numpy())
            ei_true = np.array(y_test[2].numpy())

            # ~ metrics for the total test set
            e.info('calculating prediction metrics...')
            mse_value = mean_squared_error(out_pred, out_true)
            rmse_value = np.sqrt(mse_value)
            mae_value = mean_absolute_error(out_pred, out_true)
            r2_value = r2_score(out_pred, out_true)
            e[f'mse/{rep}'] = mse_value
            e[f'rmse/{rep}'] = rmse_value
            e[f'mae/{rep}'] = mae_value
            e[f'r2/{rep}'] = r2_value

            e.info('calculating fidelity...')
            fidelities = e.apply_hook(
                'calculate_fidelity',
                model=model,
                indices_true=test_indices,
                x_true=x_test,
                y_true=y_test,
                out_pred=out_pred,
                ni_pred=ni_pred,
                ei_pred=ei_pred,
            )
            for index, fidelity in zip(test_indices, fidelities):
                e[f'fidelity/{rep}/{index}'] = fidelity

            # ~ calculating explanation metrics
            e.info('calculating explanation metrics...')

            pred_node_importances = []
            pred_edge_importances = []

            true_node_importances = []
            true_edge_importances = []

            for c, index in enumerate(test_indices):
                g = index_data_map[index]['metadata']['graph']
                ni = array_normalize(ni_pred[c])
                ei = array_normalize(ei_pred[c])

                e[f"out/true/{rep}/{index}"] = float(g['graph_labels'][0])
                e[f"out/pred/{rep}/{index}"] = float(out_pred[c][0])

                e[f'ni/pred/{rep}/{index}'] = ni
                e[f'ei/pred/{rep}/{index}'] = ei

                pred_node_importances += ni_pred[c].flatten().tolist()
                pred_edge_importances += ei_pred[c].flatten().tolist()

                true_node_importances += ni_true[c].flatten().tolist()
                true_edge_importances += ei_true[c].flatten().tolist()

                # ~ sparsity of the explanations
                node_sparsity = float(np.mean(binary_threshold(ni, 0.5)))
                edge_sparsity = float(np.mean(binary_threshold(ei, 0.5)))
                e[f'node_sparsity/{rep}/{index}'] = node_sparsity
                e[f'edge_sparsity/{rep}/{index}'] = edge_sparsity

            mean_fidelity = np.mean(list(e[f'fidelity/{rep}'].values()))
            mean_node_sparsity = np.mean(list(e[f'node_sparsity/{rep}'].values()))
            mean_edge_sparsity = np.mean(list(e[f'edge_sparsity/{rep}'].values()))

            # ~ explanation accuracy
            # In the case that there are no ground truth importances (in which case we have set them to
            # be all zeros) this will raise an exception. And in that case we will just return the worst
            # result.
            try:
                node_auc = roc_auc_score(true_node_importances, pred_node_importances)
                edge_auc = roc_auc_score(true_edge_importances, pred_edge_importances)
            except ValueError:
                node_auc = 0.5
                edge_auc = 0.5

            e[f'node_auc/{rep}'] = node_auc
            e[f'edge_auc/{rep}'] = edge_auc

            e.info(f' = mse: {mse_value:.2f}'
                   f' - rmse: {rmse_value:.2f}'
                   f' - mae: {mae_value:.2f}'
                   f' - r2: {r2_value:.2f}'
                   f' - mean fidelity: {mean_fidelity:.2f}'
                   f' - mean node sparsity: {mean_node_sparsity:.2f}'
                   f' - mean edge sparsity: {mean_edge_sparsity:.2f}'
                   f' - node auc: {node_auc:.2f}'
                   f' - edge auc: {edge_auc:.2f}')

            # -- VISUALIZATION OF RESULTS --

            # ~ visualizing examples of the graph
            e.info('drawing example visualizations...')
            # First of all we need to query the model using only the elements of the example set
            ni_example_pred = []
            ei_example_pred = []
            for index in example_indices:
                ni_example_pred.append(e[f'ni/pred/{rep}/{index}'])
                ei_example_pred.append(e[f'ei/pred/{rep}/{index}'])

            # Then we can use the already existing method "create_importances_pdf" to create a pdf file
            # which contains the visualizations of the explanations
            pdf_path = os.path.join(e.path, f'examples_{rep:02d}.pdf')
            graph_list = [index_data_map[i]['metadata']['graph'] for i in example_indices]
            image_path_list = [index_data_map[i]['image_path'] for i in example_indices]
            node_positions_list = [g['node_positions'] for g in graph_list]
            create_importances_pdf(
                graph_list=graph_list,
                image_path_list=image_path_list,
                node_positions_list=node_positions_list,
                importances_map={
                    'model': (ni_example_pred, ei_example_pred)
                },
                output_path=pdf_path,
                importance_channel_labels=IMPORTANCE_CHANNEL_LABELS,
                labels_list=[index_data_map[i]['metadata']['name'] for i in example_indices]
            )

            e.status()


with Skippable(), e.analysis:
    # Creating latex code to display the results in a table
    e.info('rendering latex table...')
    column_names = [
        r'Target Name',
        r'$\text{MSE} \downarrow $',
        r'$\text{RMSE} \downarrow $',
        r'$\text{MAE} \downarrow $',
        r'$R^2 \uparrow $',
        r'$\text{Fidelity} \uparrow$',
        r'$\text{Node Sparsity} \downarrow$',
        r'$\text{Edge Sparsity} \downarrow$',
        r'$\text{Node AUC} \uparrow$',
        r'$\text{Edge AUC} \uparrow$',
    ]
    rows = []

    row = []

    row.append('-')
    row.append([e[f'mse/{rep}'] for rep in range(REPETITIONS)])
    row.append([e[f'rmse/{rep}'] for rep in range(REPETITIONS)])
    row.append([e[f'mae/{rep}'] for rep in range(REPETITIONS)])
    row.append([e[f'r2/{rep}'] for rep in range(REPETITIONS)])
    row.append([e[f'fidelity/{rep}/{index}']
                for rep in range(REPETITIONS)
                for index in e[f'test_indices/{rep}']])
    row.append([e[f'node_sparsity/{rep}/{index}']
                for rep in range(REPETITIONS)
                for index in e[f'test_indices/{rep}']])
    row.append([e[f'edge_sparsity/{rep}/{index}']
                for rep in range(REPETITIONS)
                for index in e[f'test_indices/{rep}']])
    row.append([e[f'node_auc/{rep}'] for rep in range(REPETITIONS)])
    row.append([e[f'edge_auc/{rep}'] for rep in range(REPETITIONS)])

    rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_mean,
        caption=f'Results of {REPETITIONS} repetition(s) of ' + r'\textbf{' + MODEL_NAME + '}'
    )
    e.commit_raw('table.tex', table)
    pdf_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=pdf_path)
    e.info('rendered latex table')


"""
Inherits ""

**CHANGLELOG**

0.1.0 - xx.xx.2023 - Initial version
"""
import os
import pathlib
import typing as t

from pycomex.util import Skippable
from pycomex.experiment import SubExperiment


# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, '')
BASE_PATH = PATH
NAMESPACE = 'results/'
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):
    pass


import os
import pathlib
import typing as t

from pycomex.util import Skippable
from pycomex.experiment import SubExperiment

# == DATASET PARAMETERS ==
VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/aqsoldb')
USE_DATASET_SPLIT: t.Optional[int] = 0
TRAIN_RATIO: float = 0.8
NUM_EXAMPLES: int = 100

NODE_IMPORTANCES_KEY: t.Optional[str] = None
EDGE_IMPORTANCES_KEY: t.Optional[str] = None

USE_NODE_COORDINATES: bool = False
USE_EDGE_ATTRIBUTES: bool = True
USE_EDGE_LENGTHS: bool = False

# == MODEL PARAMETERS ==
UNITS = [32, 32, 32]
FINAL_UNITS = [64, 32, 1]
REGRESSION_REFERENCE = -3
REGRESSION_LIMITS = [[-12, 3]]

# == TRAINING PARAMETERS ==
BATCH_SIZE = 128
REPETITIONS = 5

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_single_megan.py')
BASE_PATH = PATH
NAMESPACE = 'results/vgd_single_megan_aqsoldb'
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):
    pass


"""
This example shows how to train a MEGAN model for multitask graph regression application based on an
existing "visual graph dataset".

In this example the VGD "organic_solvents" is used. This datasets consists of a few thousand molecules,
which are annotated with 4 target values, representing measurements of solubility within 4 different
organic solvents: water, ethanol, benzene, acetone. For this dataset it is important to note that not
every element is annotated with all values. Most of the time, the target value vectors contain "NaN" fields
which means that the downstream processing pipeline and graph neural network are able to handle this.

**CHANGELOG**

0.1.0 - 16.12.2022 - Initial version

0.2.0 - 30.12.2022 - Additional artifacts have been added: (1) After loading of the dataset a PDF with
information about the target value distribution of the dataset is now created. This is supposed to help
for deciding on appropriate REGRESSION_REFERENCE and REGRESSION_LIMITS values. (2) After the evaluation
step, a PDF with the test set results is created containing a regression plot and the R2 value for each of
the targets.

0.3.0 - 20.01.2023 - (1) Reworked the experiment to use the "hook" feature of the new pycomex version. The
model creating and the training are now default hook implementations which can be overwritten by
sub-experiments. (2) It is now possible to run the experiment for multiple independent REPETITIONS.
(3) The analysis code now generates a latex table which shows the average results for all the individual
targets over all the repetitions.
"""
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import csv
import random
import typing as t
from itertools import product

from pycomex.experiment import Experiment
from pycomex.util import Skippable

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from matplotlib.backends.backend_pdf import PdfPages
from kgcnn.data.moleculenet import OneHotEncoder
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from imageio.v2 import imread
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from visual_graph_datasets.util import get_dataset_path
from visual_graph_datasets.data import load_visual_graph_dataset
from visual_graph_datasets.visualization.importances import create_importances_pdf

import graph_attention_student.typing as tc
from graph_attention_student.util import latex_table, latex_table_element_mean
from graph_attention_student.util import render_latex
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.training import NoLoss, mse
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.models import Megan
from graph_attention_student.visualization import plot_regression_fit

np.set_printoptions(precision=2)

VERSION = '0.3.0'
SHORT_DESCRIPTION = (
    'This example shows how to train a MEGAN model for multitask graph regression application based on an '
    'existing visual graph dataset.'
)

# == DATASET PARAMETERS ==
# These parameters are used to specify the dataset to be used for the training as well as additional
# properties of the dataset such as the train test split for example.

# The name of the visual graph dataset to use for this experiment.
VISUAL_GRAPH_DATASET_PATH = os.path.expanduser('~/.visual_graph_datasets/datasets/organic_solvents')
# The ratio of how many elements of the dataset are supposed to be used for the training dataset.
# The rest of them will be used for the test set.
TRAIN_RATIO = 0.8
# The number of target values for each graph in the dataset.
NUM_TARGETS = 4
# Whether the dataset already contains importance (explanation) ground truth annotations.
# most of the time this will most likely not be the case
HAS_IMPORTANCES: bool = False
# IF the dataset includes global graph attributes and if they are supposed to be used in the training
# process, this flag has to be set to True.
HAS_GRAPH_ATTRIBUTES: bool = False
# The ratio of the test set to be used as examples for the visualization of the explanations
EXAMPLES_RATIO: float = 0.2
# The string names of the target values in the order in which they appear in the dataset as well
# which will be used in the labels for the result visualizations
TARGET_NAMES: t.List[str] = [
    'water',
    'benzene',
    'acetone',
    'ethanol',
]

# == MODEL PARAMETERS ==
# These paremeters can be used to configure the model

MODEL_NAME: str = 'MEGAN'

# This list defines how many graph convolutional layers to configure the network with. Every element adds
# one layer. The numbers themselves are the hidden units to be used for those layers.
UNITS: t.List[int] = [32, 32, 32, 32]

# The dropout rate which is applied after EVERY graph convolutional layer of the network. Especially for
# large networks (>20 hidden units and multiple importance channels, this dropout proves very useful)
DROPOUT_RATE: float = 0.1

# This is the weight of the additional explanation-only step which is being applied to the network.
# This explanation only step is important to develop actually interpretable explanations. Refer to the
# paper for more details about this.
IMPORTANCE_FACTOR: float = 1.0

# This is another hyperparameter of the explanation only train step. Usually should be between 1 and 10
IMPORTANCE_MULTIPLIER: float = 5

# This is the number of explanation channels that are generated by the model. This is also the number of
# attention heads used in the graph convolutional layers of the network. So to get a "wider" network, this
# parameter can also be increased. However, note that the value can only be != 2 if importance factor is
# set to exactly 0.0!
IMPORTANCE_CHANNELS: int = 2 * NUM_TARGETS

# The coefficient value of the explanation sparsity regularization that is applied to the network. Higher
# values should lead to sparser explanations.
SPARSITY_FACTOR: float = 5.0

# We need to supply the range of possible target values and a reference target value a priori to the
# network for the regression case. The regression limits should be as complete as possible. The reference
# does not have to be in the middle, changing it will influence the development of the explanations quite
# a bit.
REGRESSION_REFERENCE = [
    -3,  # water
    -1,  # benzene
    -1,  # acetone
    -1,  # ethanol
]
REGRESSION_LIMITS = [
    [-10, +2],
    [-4, +1],
    [-4, +1],
    [-4, +1],
]

# At the tail end of the network there is a MLP, which does the final prediction. This list defines the
# depth of that MLP. Each element adds a layer and the values themselves dictate the hidden values.
# In the regression case a final entry with "1" hidden layer as the output layer will implicitly be
# added.
FINAL_UNITS = [100, 50]

# This dropout is applied after EVERY layer in the final MLP. Using this should not be necessary
FINAL_DROPOUT = 0.1

# == TRAINING PARAMETERS ==
# These parameters are hyperparameters that control the training process itself.

REPETITIONS = 5
DEVICE = 'gpu:0'
OPTIMIZER_CB = lambda: ks.optimizers.Nadam(learning_rate=0.001)
BATCH_SIZE = 128
EPOCHS = 200

# == EVALUATION PARAMETERS ==
LOG_STEP = 5
LOG_STEP_EVAL = 1000
METRIC_KEY = 'mean_squared_error'
COLOR_PRIMARY = 'gray'
COLOR_SECONDARY = 'black'

# == EXPERIMENT PARAMETERS ==
DEBUG = True
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_multitask_megan'
with Skippable(), (e := Experiment(base_path=BASE_PATH, namespace=NAMESPACE, glob=globals(), debug_mode=DEBUG)):
    FINAL_UNITS += [NUM_TARGETS]

    e.info('starting experiment for VGD multitask example...')
    e.apply_hook('start_experiment')

    e.info('loading dataset...')
    name_data_map, index_data_map = load_visual_graph_dataset(
        path=VISUAL_GRAPH_DATASET_PATH,
        logger=e.logger,
        log_step=LOG_STEP_EVAL,
        metadata_contains_index=True
    )
    dataset_indices: t.List[int] = list(sorted(index_data_map.keys()))
    dataset_indices_set: t.Set[int] = set(dataset_indices)
    dataset_length = len(index_data_map)
    dataset: t.List[tc.GraphDict] = []
    for index in dataset_indices:
        data = index_data_map[index]
        g = data['metadata']['graph']
        # g['graph_labels'] = [v if v is not None else 0 for v in g['graph_labels']]
        if not HAS_IMPORTANCES:
            g['node_importances'] = np.zeros(shape=(len(g['node_indices']), IMPORTANCE_CHANNELS))
            g['edge_importances'] = np.zeros(shape=(len(g['edge_indices']), IMPORTANCE_CHANNELS))

        dataset.append(g)

    # ~ Visualize Dataset Properties
    # Here we create a PDF which illustrates some properties of the dataset relating to the target values.
    pdf_path = os.path.join(e.path, 'dataset.pdf')
    with PdfPages(pdf_path) as pdf:
        # First we are going to create a bar chart which will show for every target value, how many elements
        # are actually annotated with a valid value (which is not None)
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(12, 10))
        fig.suptitle(f'number of elements with valid target values\n'
                     f'{dataset_length} elements total')
        ax.set_xlabel('target property')
        ax.set_ylabel('number of elements')
        ax.set_xticks(range(len(TARGET_NAMES)))
        ax.set_xticklabels(TARGET_NAMES)

        target_values_map: t.Dict[str, t.List[float]] = {}
        for i, name in enumerate(TARGET_NAMES):
            target_values_map[name] = [v for g in dataset if (v := g['graph_labels'][i]) is not None]
            num_valid = len(target_values_map[name])
            ax.bar(
                x=i,
                height=num_valid,
                color=COLOR_PRIMARY
            )

        pdf.savefig(fig)
        plt.close(fig)

        # We plot the value ranges and mean values for all the different target value distributions.
        # This information will be important to decide on the model hyperparameters REGRESSION_REFERENCE
        # and REGRESSION_LIMIT for all of the targets
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(12, 10))
        fig.suptitle(f'target value ranges')
        ax.set_xlabel('target value')
        ax.set_ylabel('target property')
        ax.set_yticks(range(len(TARGET_NAMES)))

        y_labels = []
        for i, name in enumerate(TARGET_NAMES):
            values = target_values_map[name]
            min_value = np.min(values)
            max_value = np.max(values)
            mean_value = np.mean(values)
            ax.barh(
                y=i,
                left=min_value,
                width=abs(max_value - min_value),
                color=COLOR_PRIMARY,
            )
            ax.scatter(
                y=i, x=mean_value,
                color=COLOR_SECONDARY,
                marker='d',
            )
            y_labels.append(f'{name}\n'
                            f'({min_value:.1f} ; {mean_value:.1f} ; {max_value:.1f})')

        ax.set_yticklabels(y_labels)
        pdf.savefig(fig)
        plt.close(fig)

    # -- TRAINING THE MODEL

    @e.hook('create_model', default=True)
    def create_model(_e):

        model = Megan(
            units=_e.parameters['UNITS'],
            final_units=_e.parameters['FINAL_UNITS'],
            use_bias=True,
            importance_channels=_e.parameters['IMPORTANCE_CHANNELS'],
            importance_factor=_e.parameters['IMPORTANCE_FACTOR'],
            sparsity_factor=_e.parameters['SPARSITY_FACTOR'],
            regression_limits=_e.parameters['REGRESSION_LIMITS'],
            regression_reference=_e.parameters['REGRESSION_REFERENCE'],
            dropout_rate=_e.parameters['DROPOUT_RATE'],
            final_dropout_rate=_e.parameters['FINAL_DROPOUT'],
            use_graph_attributes=_e.parameters['HAS_GRAPH_ATTRIBUTES'],
        )

        model.compile(
            loss=[
                mse,
                NoLoss(),
                NoLoss()
            ],
            loss_weights=[
                1,
                0,
                0
            ],
            metrics=[
                mse,
                ks.metrics.MeanAbsoluteError(),
            ],
            optimizer=_e.parameters['OPTIMIZER_CB'](),
            run_eagerly=False
        )

        return model

    @e.hook('model_training', default=True)
    def model_training(_e, model, x_train, y_train, x_test, y_test):
        history = model.fit(
            x_train,
            y_train,
            batch_size=_e.parameters['BATCH_SIZE'],
            epochs=_e.parameters['EPOCHS'],
            validation_data=(x_test, y_test),
            validation_freq=1,
            callbacks=LogProgressCallback(
                logger=_e.logger,
                epoch_step=_e.parameters['LOG_STEP'],
                identifier=f'val_output_1_{METRIC_KEY}'
            ),
            verbose=0
        )

        return history

    for rep in range(REPETITIONS):
        e.info(f'REPETITION ({rep+1}/{REPETITIONS})')

        with tf.device(DEVICE):

            e.info('creating random train-test split...')
            train_indices = random.sample(dataset_indices, k=int(TRAIN_RATIO * len(dataset_indices)))
            train_indices_set = set(train_indices)
            test_indices_set = dataset_indices_set.difference(train_indices_set)
            test_indices = list(test_indices_set)
            e['train_indices'] = train_indices
            e['test_indices'] = test_indices
            e.info(f'chose {len(train_indices)} train indices and {len(test_indices)} test indices')

            # This turns the list of graph dicts into the final form which we need for the training
            # of the model:
            # keras RaggedTensors which contain all the graphs.
            e.info('converting dataset into ragged tensors for model training...')
            x_train, y_train, x_test, y_test = process_graph_dataset(
                dataset,
                train_indices=train_indices,
                test_indices=test_indices,
                use_graph_attributes=HAS_GRAPH_ATTRIBUTES,
            )

            e.info('creating the model...')
            model = e.apply_hook('create_model')

            e.info('starting model training...')
            history = e.apply_hook(
                'model_training',
                model=model,
                x_train=x_train,
                y_train=y_train,
                x_test=x_test,
                y_test=y_test,
            )
            e[f'history/{rep}'] = history

        # -- EVALUATION OF TEST SET --
        e.info('evaluating test set...')

        # first of all we need to pipe the entire test set through the model to get the target value
        # predictions as well as the predicted node and edge importance explanations
        out_pred, ni_pred, ei_pred = model(x_test)
        out_pred = np.array(out_pred.numpy())
        out_true = y_test[0]

        example_indices = random.sample(test_indices, k=int(EXAMPLES_RATIO * len(test_indices)))
        x_example, y_example, _, _ = process_graph_dataset(
            dataset=dataset,
            train_indices=example_indices,
            test_indices=example_indices,
            use_graph_attributes=HAS_GRAPH_ATTRIBUTES,
        )
        out_example, ni_example, ei_example = model(x_example)
        out_example = out_example.numpy()
        ni_example = ni_example.numpy()
        ei_example = ei_example.numpy()

        # -- VISUALIZATION OF RESULTS --
        # The purpose of the following code is to create artifacts which document the results of the model
        # training process in a human-readable way.

        # ~ Visualizing the evaluation results
        # This section generates a PDF which contains the R2 scores for all the different target values,
        # calculated using the test set predictions.
        # The PDF also visualizes the regression results using a regression "scatter" plot.
        pdf_path = os.path.join(e.path, f'training_{rep:02d}.pdf')
        with PdfPages(pdf_path) as pdf:
            n_cols = len(TARGET_NAMES)
            n_rows = 1

            fig, rows = plt.subplots(ncols=n_cols, nrows=n_rows, figsize=(10 * n_cols, 10), squeeze=False)
            for i, name in enumerate(TARGET_NAMES):
                ax = rows[0][i]

                values_true = []
                values_pred = []
                for value_true, value_pred in zip(out_true, out_pred):
                    if not np.isnan(value_true[i]) and not np.isnan(value_pred[i]):
                        values_true.append(value_true[i])
                        values_pred.append(value_pred[i])

                plot_regression_fit(
                    values_true=values_true,
                    values_pred=values_pred,
                    ax=ax
                )

                r2_value = r2_score(values_true, values_pred)
                mse_value = mean_squared_error(values_true, values_pred)
                rmse_value = np.sqrt(mse_value)
                ax.set_title(f'"{name}"\n'
                             f'r2: {r2_value:.2f}\n'
                             f'mse: {mse_value:.2f}\n'
                             f'rmse: {rmse_value: .2f}')

                e[f'r2/{name}/{rep}'] = r2_value
                e[f'mse/{name}/{rep}'] = mse_value
                e[f'rmse/{name}/{rep}'] = rmse_value

            pdf.savefig(fig)
            plt.close(fig)

        # ~ Visualizing Examples
        # This section will create a PDF which contains the illustrations of the predicted importance
        # explanations of the model for a subset of the test set elements contained within the randomly
        # sampled "example set".
        e.info(f'visualizing {len(example_indices)} example explanations...')
        pdf_path = os.path.join(e.path, f'examples_{rep:02d}.pdf')
        graph_list = [index_data_map[i]['metadata']['graph'] for i in example_indices]
        # "create_importances_pdf" is a rather complex function, which, if given sufficient information about
        # the dataset and the model predictions, will create a PDF file containing the visualized importance
        # explanations using graph visualizations of the underlying VGD at a target location.
        create_importances_pdf(
            graph_list=graph_list,
            image_path_list=[index_data_map[i]['image_path'] for i in example_indices],
            node_positions_list=[index_data_map[i]['metadata']['graph']['node_positions'] for i in example_indices],
            importances_map={
                'model': (ni_example, ei_example)
            },
            labels_list=[f'true {index_data_map[i]["metadata"]["target"]} - pred {out_example[c]}'
                         for c, i in enumerate(example_indices)],
            output_path=pdf_path,
            importance_channel_labels=[f'{name} {direction}'
                                       for name in TARGET_NAMES
                                       for direction in ['negative', 'positive']]
        )


with Skippable(), e.analysis:
    # Creating latex code to display the results in a table
    e.info('rendering latex table...')
    column_names = [
        r'Target Name',
        r'$\text{MSE} \downarrow $',
        r'$\text{RMSE} \downarrow $',
        r'$R^2 \uparrow $',
    ]
    rows = []
    for name in TARGET_NAMES:
        row = []

        row.append(name)
        row.append([e[f'mse/{name}/{rep}'] for rep in range(REPETITIONS)])
        row.append([e[f'rmse/{name}/{rep}'] for rep in range(REPETITIONS)])
        row.append([e[f'r2/{name}/{rep}'] for rep in range(REPETITIONS)])

        rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_mean,
        caption=f'Results of {REPETITIONS} repetition(s) of ' + r'\textbf{' + MODEL_NAME + '}'
    )
    e.commit_raw('table.tex', table)
    pdf_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=pdf_path)
    e.info('rendered latex table')


import sys
import typing as t
from typing import List

import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.ops.partition import partition_row_indexing
from kgcnn.ops.segment import segment_ops_by_name, segment_softmax
from kgcnn.layers.modules import GraphBaseLayer
from kgcnn.layers.modules import LazyConcatenate, LazyAverage
from kgcnn.layers.modules import DropoutEmbedding, DenseEmbedding, ActivationEmbedding
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.conv.gat_conv import PoolingLocalEdgesAttention
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2


class CoefficientActivation(GraphBaseLayer):

    def __init__(self,
                 activation: str = 'relu',
                 coefficient: float = 1.0,
                 ):
        super(CoefficientActivation, self).__init__()
        self.activation = activation
        self.coefficient = coefficient

        self.lay_activation = ActivationEmbedding(activation=activation)

    def call(self, inputs, *args, **kwargs):
        x = inputs
        return self.coefficient * self.lay_activation(x)


class ExtendedPoolingLocalEdgesAttention(PoolingLocalEdgesAttention):
    r"""
    """

    def __init__(self, pooling_index=0, **kwargs):
        """Initialize layer."""
        super(PoolingLocalEdgesAttention, self).__init__(**kwargs)
        self.pooling_index = pooling_index

    def call(self, inputs, **kwargs):
        """Forward pass.
        Args:
            inputs: [node, edges, attention, edge_indices]
                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
                - attention (tf.RaggedTensor): Attention coefficients of shape (batch, [M], 1)
                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], F)
        Returns:
            tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node of shape (batch, [N], F)
        """
        # Need ragged input but can be generalized in the future.
        self.assert_ragged_input_rank(inputs)
        # We cast to values here
        nod, node_part = inputs[0].values, inputs[0].row_lengths()
        edge = inputs[1].values
        attention = inputs[2].values
        edgeind, edge_part = inputs[3].values, inputs[3].row_lengths()

        shiftind = partition_row_indexing(edgeind, node_part, edge_part, partition_type_target="row_length",
                                          partition_type_index="row_length", to_indexing='batch',
                                          from_indexing=self.node_indexing)

        nodind = shiftind[:, self.pooling_index]  # Pick first index eg. ingoing
        dens = edge
        ats = attention
        if not self.is_sorted:
            # Sort edgeindices
            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
            nodind = tf.gather(nodind, node_order, axis=0)
            dens = tf.gather(dens, node_order, axis=0)
            ats = tf.gather(ats, node_order, axis=0)

        # Apply segmented softmax
        ats = segment_softmax(ats, nodind)
        get = dens * ats
        get = tf.math.segment_sum(get, nodind)

        if self.has_unconnected:
            # Need to fill tensor since the maximum node may not be also in pooled
            # Does not happen if all nodes are also connected
            get = tf.scatter_nd(tf.keras.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))

        out = tf.RaggedTensor.from_row_lengths(get, node_part, validate=self.ragged_validate)
        ats = tf.RaggedTensor.from_row_lengths(ats, node_part, validate=self.ragged_validate)

        return out, ats

    def get_config(self):
        """Update layer config."""
        config = super(PoolingLocalEdgesAttention, self).get_config()
        config.update({"pooling_index": self.pooling_index})
        return config


class MultiChannelGatLayer(AttentionHeadGATV2):
    """
    This class is mostly a copy and modification of ``kgcnn.layers.conv.attention.AttentionHeadGATV2``.
    """
    def __init__(self,
                 *args,
                 units: int,
                 activation: str = 'kgcnn>leaky_relu',
                 use_bias: bool = True,
                 channels: int = 1,
                 attention_dropout_rate: float = 0,
                 share_weights: bool = False,
                 **kwargs):
        super(MultiChannelGatLayer, self).__init__(
            *args,
            units=units,
            activation=activation,
            use_bias=use_bias,
            **kwargs
        )

        self.attention_dropout_rate = attention_dropout_rate
        self.channels = channels
        self.lay_dropout = DropoutEmbedding(rate=attention_dropout_rate)

        if share_weights:
            lay_linear = DenseEmbedding(units, activation='linear', use_bias=use_bias)

        self.alpha_layers = []
        for i in range(channels):
            if not share_weights:
                lay_linear = DenseEmbedding(units, activation='linear', use_bias=use_bias)

            lay_alpha_activation = DenseEmbedding(units, activation=activation, use_bias=use_bias)
            lay_alpha = DenseEmbedding(1, activation="linear", use_bias=False)
            self.alpha_layers.append((lay_linear, lay_alpha_activation, lay_alpha))

        self.lay_concat_alphas = LazyConcatenate(axis=-2)
        self.lay_concat_embeddings = LazyConcatenate(axis=-2)
        self.lay_pool_attention = ExtendedPoolingLocalEdgesAttention()

    def call(self, inputs, **kwargs):
        """Forward pass.

        N: Number of nodes in the graph
        V: Number of embeddings per node for the previous layer.
        K: Number of channels

        M: Number of edges in the graph
        F: Number of edge embeddings

        Args:
            inputs (list): of [nodes, edges, edge_indices]
                - nodes (tf.RaggedTensor): Node embeddings of shape ([batch], [N], K, V)
                - edges (tf.RaggedTensor): Edge or message embeddings of shape ([batch], [M], F)
                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape ([batch], [M], 2)
        Returns:
            tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node.
        """
        nodes, edge, edge_index = inputs

        # "a_ij" is a single-channel edge attention coefficient tensor. "h_i" is a single-channel node
        # embedding vector. These lists collect the single tensors for each channel and at the end are
        # concat into the multi-channel tensors.
        a_ijs = []
        h_is = []
        for k, (lay_linear, lay_alpha_activation, lay_alpha) in enumerate(self.alpha_layers):

            node = nodes[:, :, k, :]

            # Copied from the original class
            w_n = lay_linear(node, **kwargs)
            n_in = self.lay_gather_in([node, edge_index], **kwargs)
            n_out = self.lay_gather_out([node, edge_index], **kwargs)
            wn_out = self.lay_gather_out([w_n, edge_index], **kwargs)
            if self.use_edge_features:
                e_ij = self.lay_concat([n_in, n_out, edge], **kwargs)
            else:
                e_ij = self.lay_concat([n_in, n_out], **kwargs)

            a_ij = lay_alpha_activation(e_ij, **kwargs)
            a_ij = lay_alpha(a_ij, **kwargs)

            # Added to hopefully improve overfit problem
            a_ij = self.lay_dropout(a_ij)

            h_i, ats_ij = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)
            #print(a_ij.shape, ats_ij.shape)

            if self.use_final_activation:
                h_i = self.lay_final_activ(h_i, **kwargs)

            #a_ij = ks.activations.softmax(a_ij, axis=-2)
            a_ij = tf.expand_dims(a_ij, axis=-2)
            a_ijs.append(a_ij)
            # ats_ij = tf.expand_dims(ats_ij, axis=-2)
            # a_ijs.append(ats_ij)

            h_i = tf.expand_dims(h_i, axis=-2)
            h_is.append(h_i)

        a_ijs = self.lay_concat_alphas(a_ijs)
        h_is = self.lay_concat_embeddings(h_is)

        # An important modification we need here is that this layer also returns the attention coefficients
        # because we calculate the edge importances from those.
        # h_is: ([batch], [N], K, V)
        # a_ijs: ([batch], [M], K, 1)
        return h_is, a_ijs


class MultiHeadGatLayer(AttentionHeadGATV2):
    """
    This class is mostly a copy and modification of ``kgcnn.layers.conv.attention.AttentionHeadGATV2``.
    """
    def __init__(self,
                 *args,
                 units: int,
                 num_heads: int,
                 activation: str = 'kgcnn>leaky_relu',
                 use_bias: bool = True,
                 concat_heads: bool = False,
                 attention_dropout_rate: float = 0,
                 share_weights: bool = False,
                 **kwargs):
        super(MultiHeadGatLayer, self).__init__(
            *args,
            units=units,
            activation=activation,
            use_bias=use_bias,
            **kwargs
        )

        self.attention_dropout_rate = attention_dropout_rate
        self.num_heads = num_heads
        self.concat_heads = concat_heads
        self.lay_dropout = DropoutEmbedding(rate=attention_dropout_rate)

        if share_weights:
            lay_linear = DenseEmbedding(units, activation='linear', use_bias=use_bias)

        self.head_layers: t.List[t.Tuple[DenseEmbedding, DenseEmbedding, DenseEmbedding]] = []
        for _ in range(num_heads):
            if not share_weights:
                lay_linear = DenseEmbedding(units, activation='linear', use_bias=use_bias)

            lay_alpha_activation = DenseEmbedding(units, activation=activation, use_bias=use_bias)
            lay_alpha = DenseEmbedding(1, activation="linear", use_bias=False)

            # Adding all the layers as a tuple to the list as representing one of the heads
            self.head_layers.append((lay_linear, lay_alpha_activation, lay_alpha))

        self.lay_concat_alphas = LazyConcatenate(axis=-2)
        self.lay_concat_embeddings = LazyConcatenate(axis=-2)
        self.lay_pool_attention = PoolingLocalEdgesAttention()
        self.lay_pool = PoolingLocalEdges()

        self.lay_average_heads = LazyAverage()
        self.lay_concat_heads = LazyConcatenate(axis=-1)

    def call(self, inputs, **kwargs):
        """Forward pass.

        N: Number of nodes in the graph
        V: Number of embeddings per node for the previous layer.
        Vu: Number of embeddings per node after the this layer
        K: Number of attention heads

        M: Number of edges in the graph
        F: Number of edge embeddings

        Brackets [] indicate a ragged dimension

        Args:
            inputs (list): of [nodes, edges, edge_indices]
                - nodes (tf.RaggedTensor): Node embeddings of shape ([batch], [N], V)
                - edges (tf.RaggedTensor): Edge or message embeddings of shape ([batch], [M], F)
                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape ([batch], [M], 2)
        Returns:
            Tuple consisting of:
                - nodes (tf.RaggedTensor): Node embeddings of shape ([batch], [N], Vu * K) if head concat
                  is used and ([batch], [N], Vu) if it is not used and heads are averaged instead
                - attention coefficients (tf.RaggedTensor): The attention coefficients for the graph edges
                  of the various attention heads of shape ([batch], [M], K)
        """
        node, edge, edge_index = inputs

        # "a_ij" is a single-channel edge attention coefficient tensor. "h_i" is a single-channel node
        # embedding vector. These lists collect the single tensors for each channel and at the end are
        # concat into the multi-channel tensors.
        a_ijs = []
        h_is = []
        for k, (lay_linear, lay_alpha_activation, lay_alpha) in enumerate(self.head_layers):

            # Copied from the original class
            w_n = lay_linear(node, **kwargs)
            n_in = self.lay_gather_in([node, edge_index], **kwargs)
            n_out = self.lay_gather_out([node, edge_index], **kwargs)
            wn_out = self.lay_gather_out([w_n, edge_index], **kwargs)
            if self.use_edge_features:
                e_ij = self.lay_concat([n_in, n_out, edge], **kwargs)
            else:
                e_ij = self.lay_concat([n_in, n_out], **kwargs)

            # a_ij: ([batch], [M], 1)
            a_ij = lay_alpha_activation(e_ij, **kwargs)
            a_ij = lay_alpha(a_ij, **kwargs)

            # Added to hopefully improve overfit problem
            a_ij = self.lay_dropout(a_ij)

            # This is my own attempt at implementing an attention pooling operation. The problem with the
            # kgcnn PoolingLocalEdgesAttention is that it uses attention values which are not [0, 1] range
            # and as far as I can see there is also no way to get these actual coefficients out of the
            # process either.
            # I think this should work, because I realized that softmax should work on a ragged dimension as
            # well (at least it did not cause errors till now).
            # a_ij = ks.activations.softmax(a_ij, axis=-2)
            # h_i = self.lay_pool([node, wn_out * a_ij, edge_index], **kwargs)

            h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)
            #a_ij = ks.activations.softmax(a_ij, axis=-2)

            if self.use_final_activation:
                h_i = self.lay_final_activ(h_i, **kwargs)

            a_ij = tf.expand_dims(a_ij, axis=-2)
            a_ijs.append(a_ij)

            #h_i = tf.expand_dims(h_i, axis=-2)
            h_is.append(h_i)

        a_ijs = self.lay_concat_alphas(a_ijs)

        if self.concat_heads:
            h_is = self.lay_concat_heads(h_is)
        else:
            h_is = self.lay_average_heads(h_is)

        # An important modification we need here is that this layer also returns the attention coefficients
        # because we calculate the edge importances from those.
        # h_is: ([batch], [N], K * Vu) or ([batch], [N], Vu)
        # a_ijs: ([batch], [M], K, 1)
        return h_is, a_ijs


class ExplanationSparsityRegularization(GraphBaseLayer):

    def __init__(self,
                 coef: float = 1.0,
                 factor: t.Optional[float] = None,
                 **kwargs):
        super(ExplanationSparsityRegularization, self).__init__(**kwargs)
        self.factor = coef
        if factor is not None:
            self.factor = factor

    def call(self, inputs):
        # importances: ([batch], [N], K)
        importances = inputs

        loss = tf.reduce_mean(tf.math.abs(importances))
        self.add_loss(loss * self.factor)


class ExplanationExclusivityRegularization(GraphBaseLayer):

    def __init__(self, coef: float = 1.0, **kwargs):
        super(ExplanationExclusivityRegularization, self).__init__(**kwargs)
        self.coef = coef

    def call(self, inputs):
        # importances: ([batch], [N], K)
        importances = inputs

        # This loss function encourages the sum of all the elements along the importance axis to be =1.
        # This means that every element needs to be explained in some way.
        # It also discourages this sum of 1 to be reached by various small values and rather encourages
        # one item being =1 while the others are 0 - aka an explanation to be exclusive to one channel.
        #square_part = tf.square(tf.reduce_sum(tf.sqrt(importances + 0.1), axis=-1) - 1)
        #sqrt_part = - tf.sqrt(tf.reduce_sum(importances, axis=-1) + 0.1)
        #loss = square_part - sqrt_part

        mean = tf.reduce_mean(importances, axis=-1, keepdims=True)
        mean = tf.concat([mean for _ in range(importances.shape[-1])], axis=-1)
        loss = tf.reduce_sum(tf.square(importances - mean), axis=-1)
        loss = tf.reduce_mean(loss)
        self.add_loss(-self.coef * loss)


class StaticMultiplicationEmbedding(GraphBaseLayer):

    def __init__(self,
                 values: List[float],
                 bias: float = 0.0,
                 **kwargs):
        GraphBaseLayer.__init__(self)
        self.values = tf.convert_to_tensor(values, dtype=tf.float32)
        self.bias = bias

    def call(self, inputs):
        return tf.reduce_sum(inputs * self.values, axis=-1) + self.bias


class MultiHeadGATV2Layer(AttentionHeadGATV2):

    def __init__(self,
                 units: int,
                 num_heads: int,
                 activation: str = 'kgcnn>leaky_relu',
                 use_bias: bool = True,
                 concat_heads: bool = True,
                 **kwargs):
        super(MultiHeadGATV2Layer, self).__init__(
            units=units,
            activation=activation,
            use_bias=use_bias,
            **kwargs
        )
        self.num_heads = num_heads
        self.concat_heads = concat_heads

        self.head_layers = []
        for _ in range(num_heads):
            lay_linear = DenseEmbedding(units, activation=activation, use_bias=use_bias)
            lay_alpha_activation = DenseEmbedding(units, activation=activation, use_bias=use_bias)
            lay_alpha = DenseEmbedding(1, activation='linear', use_bias=False)

            self.head_layers.append((lay_linear, lay_alpha_activation, lay_alpha))

        self.lay_concat_alphas = LazyConcatenate(axis=-2)
        self.lay_concat_embeddings = LazyConcatenate(axis=-2)
        self.lay_pool_attention = PoolingLocalEdgesAttention()
        # self.lay_pool = PoolingLocalEdges()

        if self.concat_heads:
            self.lay_combine_heads = LazyConcatenate(axis=-1)
        else:
            self.lay_combine_heads = LazyAverage()

    def __call__(self, inputs, **kwargs):
        node, edge, edge_index = inputs

        # "a_ij" is a single-channel edge attention logits tensor. "a_ijs" is consequently the list which
        # stores these tensors for each attention head.
        # "h_i" is a single-channel node embedding tensor. "h_is" is consequently the list which stores
        # these tensors for each attention head.
        a_ijs = []
        h_is = []
        for k, (lay_linear, lay_alpha_activation, lay_alpha) in enumerate(self.head_layers):

            # Copied from the original class
            w_n = lay_linear(node, **kwargs)
            n_in = self.lay_gather_in([node, edge_index], **kwargs)
            n_out = self.lay_gather_out([node, edge_index], **kwargs)
            wn_out = self.lay_gather_out([w_n, edge_index], **kwargs)
            if self.use_edge_features:
                e_ij = self.lay_concat([n_in, n_out, edge], **kwargs)
            else:
                e_ij = self.lay_concat([n_in, n_out], **kwargs)

            # a_ij: ([batch], [M], 1)
            a_ij = lay_alpha_activation(e_ij, **kwargs)
            a_ij = lay_alpha(a_ij, **kwargs)

            # h_i: ([batch], [N], F)
            h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index], **kwargs)

            if self.use_final_activation:
                h_i = self.lay_final_activ(h_i, **kwargs)

            # a_ij after expand: ([batch], [M], 1, 1)
            a_ij = tf.expand_dims(a_ij, axis=-2)
            a_ijs.append(a_ij)

            # h_i = tf.expand_dims(h_i, axis=-2)
            h_is.append(h_i)

        a_ijs = self.lay_concat_alphas(a_ijs)

        h_is = self.lay_combine_heads(h_is)

        # An important modification we need here is that this layer also returns the attention coefficients
        # because in MEGAN we need those to calculate the edge attention values with!
        # h_is: ([batch], [N], K * Vu) or ([batch], [N], Vu)
        # a_ijs: ([batch], [M], K, 1)
        return h_is, a_ijs

    def get_config(self):
        """Update layer config."""
        config = super(MultiHeadGATV2Layer, self).get_config()
        config.update({
            'num_heads': self.num_heads,
            'concat_heads': self.concat_heads
        })

        return config



"""
Base experiment for training and comparing of multiple explanation supervised models specifically on graph
regression datasets.

This experiment will run multiple overall repetitions of model training processes to create a statistical
sample of the results. In each repetition multiple models are trained in an explanation supervised fashion
and then evaluated with respect to their prediction performance and their explanation accuracy towards the
ground truth explanations as well as the fidelity of their explanations.

**CHANGELOG**

0.1.0 - 18.03.2023 - Initial version
"""
import os
import pathlib
import random
import warnings
import typing as t

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from pycomex.util import Skippable
from pycomex.experiment import Experiment
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import r2_score
from sklearn.metrics import roc_auc_score
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from visual_graph_datasets.data import load_visual_graph_dataset
from visual_graph_datasets.visualization.importances import create_importances_pdf
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.models import Megan
from graph_attention_student.models import grad_importances
from graph_attention_student.models import GnesGradientModel
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.training import NoLoss
from graph_attention_student.training import ExplanationLoss
from graph_attention_student.training import mae as mae_loss
from graph_attention_student.util import array_normalize
from graph_attention_student.util import binary_threshold
from graph_attention_student.util import render_latex
from graph_attention_student.util import latex_table
from graph_attention_student.util import latex_table_element_mean

warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
PATH = pathlib.Path(__file__).parent.absolute()
SHORT_DESCRIPTION = (
    'Base experiment for training and comparing of multiple explanation supervised models specifically '
    'on graph regression datasets.'
)

# == DATASET PARAMETERS ==
VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/rb_dual_motifs')
RATIO_DATASET: t.Optional[float] = None
RATIO_TRAIN: float = 0.8
NODE_IMPORTANCES_KEY = 'node_importances_1'
EDGE_IMPORTANCES_KEY = 'edge_importances_1'
NUM_EXAMPLE: int = 100

# == TRAINING PARAMETERS ==
REPETITIONS = 25
DEVICE = 'cpu:0'
BATCH_SIZE = 32
OPTIMIZER_CB = lambda: ks.optimizers.Adam(learning_rate=0.01)
EPOCHS = 25

# == MODEL PARAMETERS ==
UNITS: t.List[int] = [32, 32, 32]
FINAL_UNITS: t.List[int] = [16, 1]
FINAL_ACTIVATION: str = 'linear'
SPARSITY_FACTOR: float = 1.0

MEGAN_UNITS: t.List[int] = [12, 12, 12]
MEGAN_FINAL_UNITS: t.List[int] = [6, 1]
MEGAN_IMPORTANCE_CHANNELS: int = 1

# == EVALUATION PARAMETERS ==
BATCH_SIZE_EVAL = 128
# :param NUM_RANDOM_FIDELITY: This integer number determines the number of times which the "random fidelity"
#       is going to be sampled for each element of the dataset.
NUM_RANDOM_FIDELITY: int = 10

# == EXPERIMENT PARAMETERS ==
BASE_PATH = os.path.join(PATH, 'results')
NAMESPACE = 'vgd_regression_supervised'
DEBUG = True
TESTING = False

# If the TESTING flag is set that means that an execution of the experiment is supposed to be only a
# test of functionality. In that case we want to modify several parameters such that the length of the
# experiment is drastically reduced.
if TESTING:
    RATIO_DATASET = 0.2
    REPETITIONS = 1
    EPOCHS = 10
    NUM_RANDOM_FIDELITY = 1
    NUM_EXAMPLE = 20

with Skippable(), (e := Experiment(BASE_PATH, NAMESPACE, globals())):
    vgd_name = os.path.basename(VISUAL_GRAPH_DATASET_PATH)
    e.info(f'using visual graph dataset "{vgd_name}"')

    # ~ loading the dataset
    e.info('loading visual graph dataset from disk')
    metadata_map, index_data_map = load_visual_graph_dataset(
        VISUAL_GRAPH_DATASET_PATH,
        logger=e.logger,
        log_step=1000,
        metadata_contains_index=True,
    )
    dataset_indices_all = list(sorted(index_data_map.keys()))
    dataset_length = len(dataset_indices_all)

    dataset: t.List[dict] = [None for _ in dataset_indices_all]
    for index in dataset_indices_all:
        g = index_data_map[index]['metadata']['graph']

        g['node_importances'] = np.array(g[NODE_IMPORTANCES_KEY]).astype(float)
        g['edge_importances'] = np.array(g[EDGE_IMPORTANCES_KEY]).astype(float)

        # We explicitly cast the attribute vectors as a float type as that is important for certain models
        # to work properly down the line
        g['node_attributes'] = np.array(g['node_attributes']).astype(float)
        g['edge_attributes'] = np.array(g['edge_attributes']).astype(float)

        dataset[index] = g

    e.info(f'loaded dataset with {len(dataset)} elements')

    @e.hook('model_generator')
    def model_generator(_e):

        # ~ fixed GNES version
        # This versions uses the absolute activation, which works better for regression problems
        model = GnesGradientModel(
            units=_e.parameters['UNITS'],
            batch_size=_e.parameters['BATCH_SIZE'],
            final_units=_e.parameters['FINAL_UNITS'],
            final_activation=_e.parameters['FINAL_ACTIVATION'],
            layer_cb=lambda units: AttentionHeadGATV2(
                units=units,
                use_edge_features=True,
            ),
            importance_func=lambda gradient_info: grad_importances(
                gradient_info=gradient_info,
                use_absolute=True,
                use_relu=False,
                keepdims=False
            ),
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(loss_function=mae_loss),
                ExplanationLoss(loss_function=mae_loss),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'gnes_fixed', model

        # ~ original GNES version
        # This versions uses the relu activation, which was proposed in the original paper but which
        # does not work very well for a regression setting.
        model = GnesGradientModel(
            units=_e.parameters['UNITS'],
            batch_size=_e.parameters['BATCH_SIZE'],
            final_units=_e.parameters['FINAL_UNITS'],
            final_activation=_e.parameters['FINAL_ACTIVATION'],
            layer_cb=lambda units: AttentionHeadGATV2(
                units=units,
                use_edge_features=True,
            ),
            importance_func=lambda gradient_info: grad_importances(
                gradient_info=gradient_info,
                use_absolute=False,
                use_relu=True,
                keepdims=False
            ),
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(loss_function=mae_loss),
                ExplanationLoss(loss_function=mae_loss),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'gnes_original', model

        # ~ MEGAN unsupervised
        # This MEGAN version does not receive explanation supervision
        model = Megan(
            units=MEGAN_UNITS,
            importance_channels=MEGAN_IMPORTANCE_CHANNELS,
            importance_factor=0.0,
            sparsity_factor=SPARSITY_FACTOR,
            concat_heads=False,
            final_units=MEGAN_FINAL_UNITS,
            final_activation=FINAL_ACTIVATION,
            use_graph_attributes=False
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                NoLoss(),
                NoLoss(),
            ],
            loss_weights=[1, 0, 0],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'megan_normal', model

        # ~ MEGAN supervised
        # This MEGAN version does not receive explanation supervision
        model = Megan(
            units=MEGAN_UNITS,
            importance_channels=MEGAN_IMPORTANCE_CHANNELS,
            importance_factor=0.0,
            sparsity_factor=SPARSITY_FACTOR,
            concat_heads=False,
            final_units=MEGAN_FINAL_UNITS,
            final_activation=FINAL_ACTIVATION,
            use_graph_attributes=False
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(),
                ExplanationLoss(),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'megan_supervised', model

    # -- MAIN LOOP --
    e.info(f'starting repetition loop...')
    for rep in range(REPETITIONS):
        e.info(f'REPETITION ({rep+1}/{REPETITIONS})')

        with tf.device(DEVICE):
            e.info('creating rep archive folder...')
            rep_path = os.path.join(e.path, f'{rep:02d}')
            os.mkdir(rep_path)

            e.info('creating random dataset split...')
            dataset_indices = dataset_indices_all.copy()
            if RATIO_DATASET:
                num_elements = int(RATIO_DATASET * dataset_length)
                dataset_indices = dataset_indices[:num_elements]

            dataset_indices_set = set(dataset_indices)
            num_train = int(len(dataset_indices) * RATIO_TRAIN)
            train_indices = random.sample(dataset_indices, k=num_train)
            train_indices_set = set(train_indices)
            test_indices_set = dataset_indices_set.difference(train_indices_set)
            test_indices = list(test_indices_set)

            num_example = min(len(test_indices), NUM_EXAMPLE)
            example_indices = random.sample(test_indices, k=num_example)

            e[f'train_indices/{rep}'] = train_indices
            e[f'test_indices/{rep}'] = test_indices
            e[f'example_indices/{rep}'] = example_indices
            e.info(f'created split with '
                   f'{len(train_indices)} train elements, '
                   f'{len(test_indices)} test elements, '
                   f'{len(example_indices)} example elements for visualization')

            e.info('processing the dataset into tensors...')
            x_train, y_train, x_test, y_test = process_graph_dataset(
                dataset,
                test_indices=test_indices,
                train_indices=train_indices,
                use_importances=True,
                use_graph_attributes=False,
            )
            e['keys'] = []
            for key, model in e.apply_hook('model_generator'):
                e.info(f'creating the model "{key}"...')
                # Here we update the current repetition index and model key into the experiment storage
                # so that this can be accessed by sub experiment hook implementations without needing
                # to explicitly pass that information every time.
                e['rep'] = rep
                e['key'] = key
                e['keys'].append(key)

                e.info('starting model training...')
                hist = model.fit(
                    x_train,
                    y_train,
                    batch_size=BATCH_SIZE,
                    epochs=EPOCHS,
                    callbacks=[LogProgressCallback(
                        logger=e.logger,
                        epoch_step=1,
                        identifier=f'loss'
                    )],
                    verbose=0,
                )
                history = hist.history
                e[f'history/{rep}'] = history
                e.info(f'finished training for model with {model.count_params()} parameters')

                e.info('evaluating on test set...')

                # ~ The overall prediction metrics
                out_pred, ni_pred, ei_pred = [v.numpy()
                                              for v in model(x_test, batch_size=len(test_indices))]
                out_true, ni_true, ei_true = [v.numpy() if not isinstance(v, np.ndarray) else v
                                              for v in y_test]

                mse = mean_squared_error(out_true, out_pred)
                mae = mean_absolute_error(out_true, out_pred)
                r2 = r2_score(out_true, out_pred)
                e.info(f'prediction metrics:\n'
                       f' * mse: {mse:.3f}\n'
                       f' * mae: {mae:.3f}\n'
                       f' * r2: {r2:.3f}\n')

                e[f'mse/{key}/{rep}'] = mse
                e[f'mae/{key}/{rep}'] = mae
                e[f'r2/{key}/{rep}'] = r2

                for c, index in enumerate(test_indices):
                    e[f'out/true/{rep}/{index}'] = out_true[c]
                    e[f'out/pred/{key}/{rep}/{index}'] = out_pred[c]

                node_masks = []
                edge_masks = []
                for c, index in enumerate(test_indices):
                    ni = array_normalize(ni_pred[c])
                    ei = array_normalize(ei_pred[c])
                    e[f'ni/{key}/{rep}/{index}'] = ni
                    e[f'ei/{key}/{rep}/{index}'] = ei

                    # Up until this point, every importance value is still a float, but we need to the binary
                    # explanations for two reasons: To construct the masks for the fidelity computation and
                    # for the computation of the sparsity
                    ni_binary = binary_threshold(ni, threshold=0.5)
                    ei_binary = binary_threshold(ei, threshold=0.5)
                    node_masks.append(1.0 - ni_binary.astype(float))
                    edge_masks.append(1.0 - ei_binary.astype(float))

                    node_sparsity = np.mean(ni_binary)
                    edge_sparsity = np.mean(ei_binary)
                    e[f'node_sparsity/{key}/{rep}/{index}'] = node_sparsity
                    e[f'edge_sparsity/{key}/{rep}/{index}'] = edge_sparsity

                e.info('calculating fidelity...')

                # Why do we sum here as well?
                # Technically the last dimension of the masks will be "1" anyway so the sum reduce has
                # absolutely no effect. We do this to eventually support multi channel explanations, which
                # have a last dimension != 1, as well, so that they don't cause an error at this point.
                node_mask_tensor = ragged_tensor_from_nested_numpy(node_masks)
                node_mask_tensor = tf.reduce_sum(node_mask_tensor, axis=-1, keepdims=True)
                edge_mask_tensor = ragged_tensor_from_nested_numpy(edge_masks)
                edge_mask_tensor = tf.reduce_sum(edge_mask_tensor, axis=-1, keepdims=True)

                out_masked, _, _ = model((
                    x_test[0] * node_mask_tensor,
                    x_test[1] * edge_mask_tensor,
                    x_test[2]
                ), create_gradients=False)
                for c, index in enumerate(test_indices):
                    # Here we calculate the fidelity as the absolute difference in the predicted value.
                    fidelity = abs(out_masked[c] - out_pred[c])
                    e[f'fidelity/{key}/{rep}/{index}'] = fidelity

                e.info('calculating random fidelity...')
                for j in range(NUM_RANDOM_FIDELITY):
                    # first of all we need to create the random masks, which we can very easily do by simply
                    # using the "correct" explanation masks we already have and simply randomly permutating
                    # them, because one important property for the random masks is that they have to have
                    # the same sparsity!
                    for node_mask, edge_mask in zip(node_masks, edge_masks):
                        np.random.shuffle(node_mask)
                        np.random.shuffle(edge_mask)

                    node_mask_tensor = ragged_tensor_from_nested_numpy(node_masks)
                    node_mask_tensor = tf.reduce_sum(node_mask_tensor, axis=-1, keepdims=True)
                    edge_mask_tensor = ragged_tensor_from_nested_numpy(edge_masks)
                    edge_mask_tensor = tf.reduce_sum(edge_mask_tensor, axis=-1, keepdims=True)

                    out_masked, _, _ = model((
                        x_test[0] * node_mask_tensor,
                        x_test[1] * edge_mask_tensor,
                        x_test[2]
                    ), create_gradients=False)
                    for c, index in enumerate(test_indices):
                        fidelity = abs(out_masked[c] - out_pred[c])  # * e[f'node_sparsity/{key}/{rep}/{index}']
                        e[f'random_fidelity/{key}/{rep}/{index}/{j}'] = fidelity

                # :hook additional_metrics:
                #       This is an action hook which can be used to inject the calculation of additional
                #       metrics, since it does not return anything to the main runtime, the results will
                #       have to be directly stored to the internal experiment storage.
                e.apply_hook(
                    'additional_metrics',
                    model=model,
                    test_indices=test_indices,
                    x_test=x_test,
                    y_test=y_test
                )

                # ~ calculating the explanation specific metrics
                node_auc = roc_auc_score(
                    np.concatenate([v.flatten() for v in ni_true]),
                    np.concatenate([v.flatten() for v in ni_pred]),
                )
                edge_auc = roc_auc_score(
                    np.concatenate([v.flatten() for v in ei_true]),
                    np.concatenate([v.flatten() for v in ei_pred]),
                )
                e[f'node_auc/{key}/{rep}'] = node_auc
                e[f'edge_auc/{key}/{rep}'] = edge_auc

                node_sparsities = [e[f'node_sparsity/{key}/{rep}/{index}'] for index in test_indices]
                edge_sparsities = [e[f'edge_sparsity/{key}/{rep}/{index}'] for index in test_indices]

                fidelities = np.array([e[f'fidelity/{key}/{rep}/{index}'] for index in test_indices])
                fidelity_mean = np.mean(fidelities)
                fidelity_std = np.std(fidelities)
                random_fidelities = np.array([np.mean([e[f'random_fidelity/{key}/{rep}/{index}/{j}']
                                                       for j in range(NUM_RANDOM_FIDELITY)])
                                              for index in test_indices])
                diff_fidelities = fidelities - random_fidelities
                for c, index in enumerate(test_indices):
                    e[f'rand_fidelity/{key}/{rep}/{index}'] = random_fidelities[c]
                    e[f'diff_fidelity/{key}/{rep}/{index}'] = diff_fidelities[c]

                e.info(f'explanation metrics:\n'
                       f' * node auc: {node_auc:.3f}\n'
                       f' * edge auc: {edge_auc:.3f}\n'
                       f' * mean node sparsity: {np.mean(node_sparsities):.3f}\n'
                       f' * mean edge sparsity: {np.mean(edge_sparsities):.3f}\n'
                       f' * mean fidelity: {fidelity_mean:.3f}\n'
                       f' * std fidelity: {fidelity_std:.3f}\n'
                       f' * mean random fidelity: {np.mean(random_fidelities):.3f}\n'
                       f' * mean diff fidelity: {np.mean(diff_fidelities):.3f}\n'
                       )

            e.info('visualizing examples...')
            graph_list = [index_data_map[i]['metadata']['graph'] for i in example_indices]
            image_path_list = [index_data_map[i]['image_path'] for i in example_indices]
            node_positions_list = [g['node_positions'] for g in graph_list]
            output_path = os.path.join(rep_path, 'examples.pdf')
            create_importances_pdf(
                graph_list=graph_list,
                image_path_list=image_path_list,
                node_positions_list=node_positions_list,
                importances_map={
                    'gt': (
                        [g['node_importances'] for g in graph_list],
                        [g['edge_importances'] for g in graph_list],
                    ),
                    **{
                        key: (
                            [e[f'ni/{key}/{rep}/{index}'] for index in example_indices],
                            [e[f'ei/{key}/{rep}/{index}'] for index in example_indices]
                        )
                        for key in e['keys']
                    }
                },
                output_path=output_path,
                logger=e.logger,
                log_step=100,
            )


with Skippable(), e.analysis:

    e.info('starting analysis of experiment results...')

    e.info('cleaning non-convergent repetitions...')
    REPETITIONS_CLEAN = []
    for rep in range(REPETITIONS):
        use_repetition = True
        for key in e['keys']:
            value = e[f'mse/{key}/{rep}']
            if value > 2.0:
                use_repetition = False

        if use_repetition:
            REPETITIONS_CLEAN.append(rep)

    e.info(f'using {len(REPETITIONS_CLEAN)} repetitions...')

    # ~ Latex table with the results over all repetitions
    e.info('rendering latex table with the results...')

    # :hook additional_columns:
    #       This hook can be used to add additional columns to the final latex evaluation table from
    #       within the child experiments. It has to return a tuple of two lists of same size.
    #       the first is a list of string names for the columns and the second list is a list of
    #       callback functions, which if given the repetition index and the key, return all the values
    #       for the corresponding metric to be used as a base for the table columns contents.
    additional_column_names, additional_callbacks = e.apply_hook('additional_columns', default=([], []))

    column_names = [
        r'Model Key',
        r'$\text{MSE} \downarrow$',
        r'$r^{2} \uparrow$',
        r'$\text{Node AUC} \uparrow$',
        r'$\text{Edge AUC} \uparrow$',
        r'$\text{Node Sparsity} \downarrow$',
        r'$\text{Edge Sparsity} \downarrow$',
        r'$\text{Fidelity} \uparrow$',
    ] + additional_column_names
    rows = []
    for key in e['keys']:
        row = []
        row.append(key.replace('_', ' '))
        row.append([e[f'mse/{key}/{rep}']
                    for rep in REPETITIONS_CLEAN])
        row.append([e[f'r2/{key}/{rep}']
                    for rep in REPETITIONS_CLEAN])
        row.append([e[f'node_auc/{key}/{rep}']
                    for rep in REPETITIONS_CLEAN])
        row.append([e[f'edge_auc/{key}/{rep}']
                    for rep in REPETITIONS_CLEAN])
        row.append([np.mean(list(e[f'node_sparsity/{key}/{rep}'].values()))
                    for rep in REPETITIONS_CLEAN])
        row.append([np.mean(list(e[f'edge_sparsity/{key}/{rep}'].values()))
                    for rep in REPETITIONS_CLEAN])
        row.append([np.mean(list(e[f'diff_fidelity/{key}/{rep}'].values()))
                    for rep in REPETITIONS_CLEAN])

        # Each one of the additional callbacks is supposed to return a list of all the values corresponding
        # to the metrics for exactly that key and that repetition
        for cb in additional_callbacks:
            row.append([np.mean(cb(key, rep)) for rep in REPETITIONS_CLEAN])

        rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_mean,
    )
    e.commit_raw('table.tex', table)
    output_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=output_path)
    e.info('rendered latex table with the results')



"""
Trains the MEGAN model on the VGD multitask TADF dataset, which uses all three values "strength",
"splitting" and "absorption" of the original TADF dataset as multitask targets.

**CHANGELOG**

0.1.0 - 20.01.2023 - Initial version
"""
import os
import pathlib

from pycomex.util import Skippable
from pycomex.experiment import SubExperiment

PATH = pathlib.Path(__file__).parent.absolute()

# == DATASET PARAMETERS ==
VISUAL_GRAPH_DATASET_PATH = os.path.expanduser('~/.visual_graph_datasets/datasets/tadf_multi')
TRAIN_RATIO = 0.6
NUM_TARGETS = 3
HAS_IMPORTANCES: bool = False
HAS_GRAPH_ATTRIBUTES: bool = True
EXAMPLES_RATIO: float = 0.01
IMPORTANCE_CHANNELS = 2 * NUM_TARGETS
TARGET_NAMES = [
    'splitting',
    'strength',
    'absorption',
]
REGRESSION_REFERENCE = [
    1,  # splitting
    1,  # strength
    3,  # absorption
]
REGRESSION_LIMITS = [
    [-1, +4],
    [-0, +5],
    [-0, +13],
]

# == TRAINING PARAMETERS ==
BATCH_SIZE = 4096
EPOCHS = 5
LOG_STEP = 1
DEVICE = 'cpu:0'

# == EXPERIMENT PARAMETERS ==
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_multitask_megan.py')
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_multitask_megan_tadf'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):
    pass



"""

"""
import os
import pathlib
import typing as t

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks
from pycomex.experiment import SubExperiment
from pycomex.util import Skippable
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from graph_attention_student.models.megan import Megan
from graph_attention_student.models.gradient import grad_importances
from graph_attention_student.models.gnes import GnesGradientModel
from graph_attention_student.training import NoLoss
from graph_attention_student.training import ExplanationLoss
from graph_attention_student.training import mae as mae_loss

PATH = pathlib.Path(__file__).parent.absolute()

# == DATASET PROPERTIES ==
VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/rb_dual_motifs')
NODE_IMPORTANCES_KEY = 'node_importances_1'
EDGE_IMPORTANCES_KEY = 'edge_importances_1'

# == MODEL PARAMETERS ==
SPARSITY_FACTOR: float = 1.0
FINAL_ACTIVATION: str = 'linear'

MEGAN_UNITS: t.List[int] = [12, 12, 12]
MEGAN_FINAL_UNITS: t.List[int] = [6, 1]
MEGAN_IMPORTANCE_CHANNELS: int = 1
MEGAN_IMPORTANCE_FACTOR: float = 0.0

GNES_UNITS: t.List[int] = [12, 12, 12]
GNES_FINAL_UNITS: t.List[int] = [6, 1]
GNES_LAYER_CB: t.Callable = lambda units: AttentionHeadGATV2(
    units=units,
    activation='kgcnn>leaky_relu',
    use_edge_features=True,
)

# == EXPERIMENT PARAMETERS ==
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_regression_supervised.py')
BASE_PATH = PATH
NAMESPACE = 'results/' + os.path.basename(__file__).strip('.py')
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    @se.hook('model_generator')
    def model_generator(e):
        # ~ fixed GNES version
        # This versions uses the absolute activation, which works better for regression problems
        model = GnesGradientModel(
            units=e.parameters['GNES_UNITS'],
            final_units=e.parameters['GNES_FINAL_UNITS'],
            batch_size=e.parameters['BATCH_SIZE'],
            final_activation=e.parameters['FINAL_ACTIVATION'],
            layer_cb=e.parameters['GNES_LAYER_CB'],
            importance_func=lambda gradient_info: grad_importances(
                gradient_info=gradient_info,
                use_absolute=True,
                use_relu=False,
                keepdims=False
            ),
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(loss_function=mae_loss),
                ExplanationLoss(loss_function=mae_loss),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'gnes_fixed', model

        # ~ original GNES version
        # This versions uses the relu activation, which was proposed in the original paper but which
        # does not work very well for a regression setting.
        model = GnesGradientModel(
            units=e.parameters['GNES_UNITS'],
            final_units=e.parameters['GNES_FINAL_UNITS'],
            batch_size=e.parameters['BATCH_SIZE'],
            final_activation=e.parameters['FINAL_ACTIVATION'],
            layer_cb=e.parameters['GNES_LAYER_CB'],
            importance_func=lambda gradient_info: grad_importances(
                gradient_info=gradient_info,
                use_absolute=False,
                use_relu=True,
                keepdims=False
            ),
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(loss_function=mae_loss),
                ExplanationLoss(loss_function=mae_loss),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'gnes_original', model

        # ~ MEGAN unsupervised
        # This MEGAN version does not receive explanation supervision
        model = Megan(
            units=e.parameters['MEGAN_UNITS'],
            final_units=e.parameters['MEGAN_FINAL_UNITS'],
            importance_channels=e.parameters['MEGAN_IMPORTANCE_CHANNELS'],
            importance_factor=0.0,
            sparsity_factor=e.parameters['SPARSITY_FACTOR'],
            final_activation=e.parameters['FINAL_ACTIVATION'],
            concat_heads=False,
            use_graph_attributes=False
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                NoLoss(),
                NoLoss(),
            ],
            loss_weights=[1, 0, 0],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'megan_1_normal', model

        # ~ MEGAN supervised
        # This MEGAN version does not receive explanation supervision
        model = Megan(
            units=e.parameters['MEGAN_UNITS'],
            final_units=e.parameters['MEGAN_FINAL_UNITS'],
            importance_channels=e.parameters['MEGAN_IMPORTANCE_CHANNELS'],
            importance_factor=0.0,
            sparsity_factor=e.parameters['SPARSITY_FACTOR'],
            final_activation=e.parameters['FINAL_ACTIVATION'],
            concat_heads=False,
            use_graph_attributes=False
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(),
                ExplanationLoss(),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )
        yield 'megan_1_supervised', model

